import os
from typing import Any, Callable, Dict, List, Optional, Tuple, cast
from unittest import mock

import altair as alt
import nbconvert
import nbformat
import pytest
from freezegun import freeze_time

from great_expectations import DataContext
from great_expectations.core import ExpectationConfiguration, ExpectationSuite
from great_expectations.core.batch import Batch
from great_expectations.core.domain import (
    INFERRED_SEMANTIC_TYPE_KEY,
    Domain,
    SemanticDomainTypes,
)
from great_expectations.core.metric_domain_types import MetricDomainTypes
from great_expectations.core.usage_statistics.events import UsageStatsEvents
from great_expectations.rule_based_profiler.altair import AltairDataTypes
from great_expectations.rule_based_profiler.config import RuleBasedProfilerConfig
from great_expectations.rule_based_profiler.data_assistant import VolumeDataAssistant
from great_expectations.rule_based_profiler.data_assistant.data_assistant_runner import (
    NumericRangeEstimatorType,
)
from great_expectations.rule_based_profiler.data_assistant_result import (
    DataAssistantResult,
    VolumeDataAssistantResult,
)
from great_expectations.rule_based_profiler.data_assistant_result.plot_result import (
    PlotResult,
)
from great_expectations.rule_based_profiler.helpers.util import (
    get_validator_with_expectation_suite,
)
from great_expectations.rule_based_profiler.parameter_container import (
    FULLY_QUALIFIED_PARAMETER_NAME_ATTRIBUTED_VALUE_KEY,
    RAW_PARAMETER_KEY,
    ParameterNode,
)
from great_expectations.util import deep_filter_properties_iterable
from great_expectations.validator.validator import Validator
from tests.render.test_util import load_notebook_from_path
from tests.test_utils import find_strings_in_nested_obj


@pytest.fixture
def quentin_expected_metrics_by_domain() -> Dict[Domain, Dict[str, Any]]:
    expected_metrics_by_domain: Dict[Domain, Dict[str, Any]] = {
        Domain(domain_type=MetricDomainTypes.TABLE, rule_name="table_rule",): {
            "$parameter.table_row_count": {
                "value": [
                    10000,
                    10000,
                    10000,
                    10000,
                    10000,
                    10000,
                    10000,
                    10000,
                    10000,
                    10000,
                    10000,
                    10000,
                    10000,
                    10000,
                    10000,
                    10000,
                    10000,
                    10000,
                    10000,
                    10000,
                    10000,
                    10000,
                    10000,
                    10000,
                    10000,
                    10000,
                    10000,
                    10000,
                    10000,
                    10000,
                    10000,
                    10000,
                    10000,
                    10000,
                    10000,
                    10000,
                ],
                "attributed_value": {
                    "c92d0679f769ac83fef2bb5eaac5d12a": [10000],
                    "562969eaef9c843cb4531aecbc13bbcb": [10000],
                    "569a4a80bf434c888593c651dbf2f157": [10000],
                    "f6c389dcef63c1f214c30f66b66945c0": [10000],
                    "c4fe9afce1cf3e83eb8518a9f5abc754": [10000],
                    "e20c38f98b9830a40b851939ca7189d4": [10000],
                    "f2e4d3da6556638b55df8ce509b094c2": [10000],
                    "44c1b1947c9049e7db62c5320dde4c63": [10000],
                    "47157bdaf05a7992473cd699cabaef74": [10000],
                    "08085632aff9ce4cebbb8023049e1aec": [10000],
                    "bb54e4fa3906387218be10cff631a7c2": [10000],
                    "58ce3b40d384eacd9bad7d916eb8f705": [10000],
                    "0327cfb13205ec8512e1c28e438ab43b": [10000],
                    "0808e185a52825d22356de2fe00a8f5f": [10000],
                    "90bb41c1fbd7c71c05dbc8695320af71": [10000],
                    "6c7e43619fe5e6963e8159cc84a28321": [10000],
                    "976b121b46db6967854b9c1a6628396b": [10000],
                    "9e58d3c72c7006b6f5800b623fbc9818": [10000],
                    "ce5f02ac408b7b5c500050190f549736": [10000],
                    "bb81456ec79522bf02f34b02762f95e0": [10000],
                    "b20800a7faafd2808d6c888577a2ba1d": [10000],
                    "33d910f95326c0c7dfe7536d1cfeba51": [10000],
                    "61e4931d87cb627df2a19b8bc5819b7b": [10000],
                    "3692b23382fd4734215465251290c65b": [10000],
                    "eff8910cddcdff62e4741243099240d5": [10000],
                    "f67d274202366f6b976414c950ca14bd": [10000],
                    "7b3ce20a8e8cf3097bb9df270a7ae63a": [10000],
                    "73612fdabd337d5a8279acc30ce22d00": [10000],
                    "ad2ad2a70c3e0bf94ddef3f893e92291": [10000],
                    "8ce0d477f610ea18e2ea4fbbb46de857": [10000],
                    "ff5a6cc031dd2c98b8bccd4766af38c1": [10000],
                    "940576153c66af14a949fd19aedd5f5b": [10000],
                    "ab05b4fb82e37c8cf5b1ac40d0a37fe9": [10000],
                    "57c04d62ada3a102248b48f34c755159": [10000],
                    "816b147dcf3305839f723a131b9ad6af": [10000],
                    "84000630d1b69a0fe870c94fb26a32bc": [10000],
                },
                "details": {
                    "metric_configuration": {
                        "metric_name": "table.row_count",
                        "domain_kwargs": {},
                        "metric_value_kwargs": None,
                    },
                    "num_batches": 36,
                },
            }
        },
        Domain(
            domain_type=MetricDomainTypes.COLUMN,
            domain_kwargs={
                "column": "passenger_count",
            },
            details={
                INFERRED_SEMANTIC_TYPE_KEY: {
                    "passenger_count": SemanticDomainTypes.NUMERIC,
                },
            },
            rule_name="categorical_columns_rule",
        ): {
            "$parameter.column_distinct_values_count": {
                "value": [
                    7,
                    7,
                    7,
                    7,
                    7,
                    7,
                    8,
                    7,
                    7,
                    7,
                    7,
                    8,
                    6,
                    7,
                    7,
                    7,
                    7,
                    8,
                    7,
                    7,
                    7,
                    7,
                    7,
                    8,
                    7,
                    7,
                    7,
                    7,
                    7,
                    7,
                    7,
                    7,
                    7,
                    7,
                    7,
                    7,
                ],
                "attributed_value": {
                    "c92d0679f769ac83fef2bb5eaac5d12a": [7],
                    "562969eaef9c843cb4531aecbc13bbcb": [7],
                    "569a4a80bf434c888593c651dbf2f157": [7],
                    "f6c389dcef63c1f214c30f66b66945c0": [7],
                    "c4fe9afce1cf3e83eb8518a9f5abc754": [7],
                    "e20c38f98b9830a40b851939ca7189d4": [7],
                    "f2e4d3da6556638b55df8ce509b094c2": [8],
                    "44c1b1947c9049e7db62c5320dde4c63": [7],
                    "47157bdaf05a7992473cd699cabaef74": [7],
                    "08085632aff9ce4cebbb8023049e1aec": [7],
                    "bb54e4fa3906387218be10cff631a7c2": [7],
                    "58ce3b40d384eacd9bad7d916eb8f705": [8],
                    "0327cfb13205ec8512e1c28e438ab43b": [6],
                    "0808e185a52825d22356de2fe00a8f5f": [7],
                    "90bb41c1fbd7c71c05dbc8695320af71": [7],
                    "6c7e43619fe5e6963e8159cc84a28321": [7],
                    "976b121b46db6967854b9c1a6628396b": [7],
                    "9e58d3c72c7006b6f5800b623fbc9818": [8],
                    "ce5f02ac408b7b5c500050190f549736": [7],
                    "bb81456ec79522bf02f34b02762f95e0": [7],
                    "b20800a7faafd2808d6c888577a2ba1d": [7],
                    "33d910f95326c0c7dfe7536d1cfeba51": [7],
                    "61e4931d87cb627df2a19b8bc5819b7b": [7],
                    "3692b23382fd4734215465251290c65b": [8],
                    "eff8910cddcdff62e4741243099240d5": [7],
                    "f67d274202366f6b976414c950ca14bd": [7],
                    "7b3ce20a8e8cf3097bb9df270a7ae63a": [7],
                    "73612fdabd337d5a8279acc30ce22d00": [7],
                    "ad2ad2a70c3e0bf94ddef3f893e92291": [7],
                    "8ce0d477f610ea18e2ea4fbbb46de857": [7],
                    "ff5a6cc031dd2c98b8bccd4766af38c1": [7],
                    "940576153c66af14a949fd19aedd5f5b": [7],
                    "ab05b4fb82e37c8cf5b1ac40d0a37fe9": [7],
                    "57c04d62ada3a102248b48f34c755159": [7],
                    "816b147dcf3305839f723a131b9ad6af": [7],
                    "84000630d1b69a0fe870c94fb26a32bc": [7],
                },
                "details": {
                    "metric_configuration": {
                        "metric_name": "column.distinct_values.count",
                        "domain_kwargs": {"column": "passenger_count"},
                        "metric_value_kwargs": None,
                    },
                    "num_batches": 36,
                },
            }
        },
        Domain(
            domain_type=MetricDomainTypes.COLUMN,
            domain_kwargs={
                "column": "store_and_fwd_flag",
            },
            details={
                INFERRED_SEMANTIC_TYPE_KEY: {
                    "store_and_fwd_flag": SemanticDomainTypes.TEXT,
                },
            },
            rule_name="categorical_columns_rule",
        ): {
            "$parameter.column_distinct_values_count": {
                "value": [
                    2,
                    2,
                    2,
                    2,
                    2,
                    2,
                    2,
                    2,
                    2,
                    2,
                    2,
                    2,
                    2,
                    2,
                    2,
                    2,
                    2,
                    2,
                    2,
                    2,
                    2,
                    2,
                    2,
                    2,
                    2,
                    2,
                    2,
                    2,
                    2,
                    2,
                    2,
                    2,
                    2,
                    2,
                    2,
                    2,
                ],
                "attributed_value": {
                    "c92d0679f769ac83fef2bb5eaac5d12a": [2],
                    "562969eaef9c843cb4531aecbc13bbcb": [2],
                    "569a4a80bf434c888593c651dbf2f157": [2],
                    "f6c389dcef63c1f214c30f66b66945c0": [2],
                    "c4fe9afce1cf3e83eb8518a9f5abc754": [2],
                    "e20c38f98b9830a40b851939ca7189d4": [2],
                    "f2e4d3da6556638b55df8ce509b094c2": [2],
                    "44c1b1947c9049e7db62c5320dde4c63": [2],
                    "47157bdaf05a7992473cd699cabaef74": [2],
                    "08085632aff9ce4cebbb8023049e1aec": [2],
                    "bb54e4fa3906387218be10cff631a7c2": [2],
                    "58ce3b40d384eacd9bad7d916eb8f705": [2],
                    "0327cfb13205ec8512e1c28e438ab43b": [2],
                    "0808e185a52825d22356de2fe00a8f5f": [2],
                    "90bb41c1fbd7c71c05dbc8695320af71": [2],
                    "6c7e43619fe5e6963e8159cc84a28321": [2],
                    "976b121b46db6967854b9c1a6628396b": [2],
                    "9e58d3c72c7006b6f5800b623fbc9818": [2],
                    "ce5f02ac408b7b5c500050190f549736": [2],
                    "bb81456ec79522bf02f34b02762f95e0": [2],
                    "b20800a7faafd2808d6c888577a2ba1d": [2],
                    "33d910f95326c0c7dfe7536d1cfeba51": [2],
                    "61e4931d87cb627df2a19b8bc5819b7b": [2],
                    "3692b23382fd4734215465251290c65b": [2],
                    "eff8910cddcdff62e4741243099240d5": [2],
                    "f67d274202366f6b976414c950ca14bd": [2],
                    "7b3ce20a8e8cf3097bb9df270a7ae63a": [2],
                    "73612fdabd337d5a8279acc30ce22d00": [2],
                    "ad2ad2a70c3e0bf94ddef3f893e92291": [2],
                    "8ce0d477f610ea18e2ea4fbbb46de857": [2],
                    "ff5a6cc031dd2c98b8bccd4766af38c1": [2],
                    "940576153c66af14a949fd19aedd5f5b": [2],
                    "ab05b4fb82e37c8cf5b1ac40d0a37fe9": [2],
                    "57c04d62ada3a102248b48f34c755159": [2],
                    "816b147dcf3305839f723a131b9ad6af": [2],
                    "84000630d1b69a0fe870c94fb26a32bc": [2],
                },
                "details": {
                    "metric_configuration": {
                        "metric_name": "column.distinct_values.count",
                        "domain_kwargs": {"column": "store_and_fwd_flag"},
                        "metric_value_kwargs": None,
                    },
                    "num_batches": 36,
                },
            }
        },
        Domain(
            domain_type=MetricDomainTypes.COLUMN,
            domain_kwargs={
                "column": "payment_type",
            },
            details={
                INFERRED_SEMANTIC_TYPE_KEY: {
                    "payment_type": SemanticDomainTypes.NUMERIC,
                },
            },
            rule_name="categorical_columns_rule",
        ): {
            "$parameter.column_distinct_values_count": {
                "value": [
                    4,
                    4,
                    4,
                    4,
                    4,
                    4,
                    4,
                    4,
                    4,
                    4,
                    4,
                    4,
                    4,
                    4,
                    4,
                    4,
                    4,
                    4,
                    4,
                    4,
                    4,
                    4,
                    4,
                    4,
                    4,
                    4,
                    4,
                    4,
                    4,
                    4,
                    4,
                    4,
                    4,
                    4,
                    4,
                    4,
                ],
                "attributed_value": {
                    "c92d0679f769ac83fef2bb5eaac5d12a": [4],
                    "562969eaef9c843cb4531aecbc13bbcb": [4],
                    "569a4a80bf434c888593c651dbf2f157": [4],
                    "f6c389dcef63c1f214c30f66b66945c0": [4],
                    "c4fe9afce1cf3e83eb8518a9f5abc754": [4],
                    "e20c38f98b9830a40b851939ca7189d4": [4],
                    "f2e4d3da6556638b55df8ce509b094c2": [4],
                    "44c1b1947c9049e7db62c5320dde4c63": [4],
                    "47157bdaf05a7992473cd699cabaef74": [4],
                    "08085632aff9ce4cebbb8023049e1aec": [4],
                    "bb54e4fa3906387218be10cff631a7c2": [4],
                    "58ce3b40d384eacd9bad7d916eb8f705": [4],
                    "0327cfb13205ec8512e1c28e438ab43b": [4],
                    "0808e185a52825d22356de2fe00a8f5f": [4],
                    "90bb41c1fbd7c71c05dbc8695320af71": [4],
                    "6c7e43619fe5e6963e8159cc84a28321": [4],
                    "976b121b46db6967854b9c1a6628396b": [4],
                    "9e58d3c72c7006b6f5800b623fbc9818": [4],
                    "ce5f02ac408b7b5c500050190f549736": [4],
                    "bb81456ec79522bf02f34b02762f95e0": [4],
                    "b20800a7faafd2808d6c888577a2ba1d": [4],
                    "33d910f95326c0c7dfe7536d1cfeba51": [4],
                    "61e4931d87cb627df2a19b8bc5819b7b": [4],
                    "3692b23382fd4734215465251290c65b": [4],
                    "eff8910cddcdff62e4741243099240d5": [4],
                    "f67d274202366f6b976414c950ca14bd": [4],
                    "7b3ce20a8e8cf3097bb9df270a7ae63a": [4],
                    "73612fdabd337d5a8279acc30ce22d00": [4],
                    "ad2ad2a70c3e0bf94ddef3f893e92291": [4],
                    "8ce0d477f610ea18e2ea4fbbb46de857": [4],
                    "ff5a6cc031dd2c98b8bccd4766af38c1": [4],
                    "940576153c66af14a949fd19aedd5f5b": [4],
                    "ab05b4fb82e37c8cf5b1ac40d0a37fe9": [4],
                    "57c04d62ada3a102248b48f34c755159": [4],
                    "816b147dcf3305839f723a131b9ad6af": [4],
                    "84000630d1b69a0fe870c94fb26a32bc": [4],
                },
                "details": {
                    "metric_configuration": {
                        "metric_name": "column.distinct_values.count",
                        "domain_kwargs": {"column": "payment_type"},
                        "metric_value_kwargs": None,
                    },
                    "num_batches": 36,
                },
            }
        },
        Domain(
            domain_type=MetricDomainTypes.COLUMN,
            domain_kwargs={
                "column": "extra",
            },
            details={
                INFERRED_SEMANTIC_TYPE_KEY: {
                    "extra": SemanticDomainTypes.NUMERIC,
                },
            },
            rule_name="categorical_columns_rule",
        ): {
            "$parameter.column_distinct_values_count": {
                "value": [
                    6,
                    6,
                    6,
                    5,
                    6,
                    4,
                    5,
                    7,
                    6,
                    7,
                    6,
                    6,
                    8,
                    10,
                    12,
                    12,
                    10,
                    11,
                    12,
                    14,
                    13,
                    16,
                    12,
                    13,
                    12,
                    15,
                    14,
                    10,
                    11,
                    10,
                    13,
                    12,
                    11,
                    10,
                    11,
                    10,
                ],
                "attributed_value": {
                    "c92d0679f769ac83fef2bb5eaac5d12a": [6],
                    "562969eaef9c843cb4531aecbc13bbcb": [6],
                    "569a4a80bf434c888593c651dbf2f157": [6],
                    "f6c389dcef63c1f214c30f66b66945c0": [5],
                    "c4fe9afce1cf3e83eb8518a9f5abc754": [6],
                    "e20c38f98b9830a40b851939ca7189d4": [4],
                    "f2e4d3da6556638b55df8ce509b094c2": [5],
                    "44c1b1947c9049e7db62c5320dde4c63": [7],
                    "47157bdaf05a7992473cd699cabaef74": [6],
                    "08085632aff9ce4cebbb8023049e1aec": [7],
                    "bb54e4fa3906387218be10cff631a7c2": [6],
                    "58ce3b40d384eacd9bad7d916eb8f705": [6],
                    "0327cfb13205ec8512e1c28e438ab43b": [8],
                    "0808e185a52825d22356de2fe00a8f5f": [10],
                    "90bb41c1fbd7c71c05dbc8695320af71": [12],
                    "6c7e43619fe5e6963e8159cc84a28321": [12],
                    "976b121b46db6967854b9c1a6628396b": [10],
                    "9e58d3c72c7006b6f5800b623fbc9818": [11],
                    "ce5f02ac408b7b5c500050190f549736": [12],
                    "bb81456ec79522bf02f34b02762f95e0": [14],
                    "b20800a7faafd2808d6c888577a2ba1d": [13],
                    "33d910f95326c0c7dfe7536d1cfeba51": [16],
                    "61e4931d87cb627df2a19b8bc5819b7b": [12],
                    "3692b23382fd4734215465251290c65b": [13],
                    "eff8910cddcdff62e4741243099240d5": [12],
                    "f67d274202366f6b976414c950ca14bd": [15],
                    "7b3ce20a8e8cf3097bb9df270a7ae63a": [14],
                    "73612fdabd337d5a8279acc30ce22d00": [10],
                    "ad2ad2a70c3e0bf94ddef3f893e92291": [11],
                    "8ce0d477f610ea18e2ea4fbbb46de857": [10],
                    "ff5a6cc031dd2c98b8bccd4766af38c1": [13],
                    "940576153c66af14a949fd19aedd5f5b": [12],
                    "ab05b4fb82e37c8cf5b1ac40d0a37fe9": [11],
                    "57c04d62ada3a102248b48f34c755159": [10],
                    "816b147dcf3305839f723a131b9ad6af": [11],
                    "84000630d1b69a0fe870c94fb26a32bc": [10],
                },
                "details": {
                    "metric_configuration": {
                        "metric_name": "column.distinct_values.count",
                        "domain_kwargs": {"column": "extra"},
                        "metric_value_kwargs": None,
                    },
                    "num_batches": 36,
                },
            }
        },
        Domain(
            domain_type=MetricDomainTypes.COLUMN,
            domain_kwargs={
                "column": "mta_tax",
            },
            details={
                INFERRED_SEMANTIC_TYPE_KEY: {
                    "mta_tax": SemanticDomainTypes.NUMERIC,
                },
            },
            rule_name="categorical_columns_rule",
        ): {
            "$parameter.column_distinct_values_count": {
                "value": [
                    3,
                    3,
                    3,
                    3,
                    3,
                    3,
                    3,
                    3,
                    3,
                    3,
                    3,
                    3,
                    4,
                    3,
                    3,
                    3,
                    3,
                    4,
                    3,
                    3,
                    3,
                    3,
                    4,
                    4,
                    3,
                    3,
                    3,
                    3,
                    3,
                    3,
                    3,
                    3,
                    3,
                    4,
                    4,
                    3,
                ],
                "attributed_value": {
                    "c92d0679f769ac83fef2bb5eaac5d12a": [3],
                    "562969eaef9c843cb4531aecbc13bbcb": [3],
                    "569a4a80bf434c888593c651dbf2f157": [3],
                    "f6c389dcef63c1f214c30f66b66945c0": [3],
                    "c4fe9afce1cf3e83eb8518a9f5abc754": [3],
                    "e20c38f98b9830a40b851939ca7189d4": [3],
                    "f2e4d3da6556638b55df8ce509b094c2": [3],
                    "44c1b1947c9049e7db62c5320dde4c63": [3],
                    "47157bdaf05a7992473cd699cabaef74": [3],
                    "08085632aff9ce4cebbb8023049e1aec": [3],
                    "bb54e4fa3906387218be10cff631a7c2": [3],
                    "58ce3b40d384eacd9bad7d916eb8f705": [3],
                    "0327cfb13205ec8512e1c28e438ab43b": [4],
                    "0808e185a52825d22356de2fe00a8f5f": [3],
                    "90bb41c1fbd7c71c05dbc8695320af71": [3],
                    "6c7e43619fe5e6963e8159cc84a28321": [3],
                    "976b121b46db6967854b9c1a6628396b": [3],
                    "9e58d3c72c7006b6f5800b623fbc9818": [4],
                    "ce5f02ac408b7b5c500050190f549736": [3],
                    "bb81456ec79522bf02f34b02762f95e0": [3],
                    "b20800a7faafd2808d6c888577a2ba1d": [3],
                    "33d910f95326c0c7dfe7536d1cfeba51": [3],
                    "61e4931d87cb627df2a19b8bc5819b7b": [4],
                    "3692b23382fd4734215465251290c65b": [4],
                    "eff8910cddcdff62e4741243099240d5": [3],
                    "f67d274202366f6b976414c950ca14bd": [3],
                    "7b3ce20a8e8cf3097bb9df270a7ae63a": [3],
                    "73612fdabd337d5a8279acc30ce22d00": [3],
                    "ad2ad2a70c3e0bf94ddef3f893e92291": [3],
                    "8ce0d477f610ea18e2ea4fbbb46de857": [3],
                    "ff5a6cc031dd2c98b8bccd4766af38c1": [3],
                    "940576153c66af14a949fd19aedd5f5b": [3],
                    "ab05b4fb82e37c8cf5b1ac40d0a37fe9": [3],
                    "57c04d62ada3a102248b48f34c755159": [4],
                    "816b147dcf3305839f723a131b9ad6af": [4],
                    "84000630d1b69a0fe870c94fb26a32bc": [3],
                },
                "details": {
                    "metric_configuration": {
                        "metric_name": "column.distinct_values.count",
                        "domain_kwargs": {"column": "mta_tax"},
                        "metric_value_kwargs": None,
                    },
                    "num_batches": 36,
                },
            }
        },
        Domain(
            domain_type=MetricDomainTypes.COLUMN,
            domain_kwargs={
                "column": "tolls_amount",
            },
            details={
                INFERRED_SEMANTIC_TYPE_KEY: {
                    "tolls_amount": SemanticDomainTypes.NUMERIC,
                },
            },
            rule_name="categorical_columns_rule",
        ): {
            "$parameter.column_distinct_values_count": {
                "value": [
                    20,
                    24,
                    28,
                    24,
                    21,
                    25,
                    23,
                    26,
                    24,
                    19,
                    24,
                    26,
                    22,
                    26,
                    20,
                    27,
                    27,
                    23,
                    31,
                    27,
                    28,
                    32,
                    23,
                    26,
                    27,
                    29,
                    20,
                    19,
                    22,
                    30,
                    27,
                    22,
                    21,
                    22,
                    16,
                    20,
                ],
                "attributed_value": {
                    "c92d0679f769ac83fef2bb5eaac5d12a": [20],
                    "562969eaef9c843cb4531aecbc13bbcb": [24],
                    "569a4a80bf434c888593c651dbf2f157": [28],
                    "f6c389dcef63c1f214c30f66b66945c0": [24],
                    "c4fe9afce1cf3e83eb8518a9f5abc754": [21],
                    "e20c38f98b9830a40b851939ca7189d4": [25],
                    "f2e4d3da6556638b55df8ce509b094c2": [23],
                    "44c1b1947c9049e7db62c5320dde4c63": [26],
                    "47157bdaf05a7992473cd699cabaef74": [24],
                    "08085632aff9ce4cebbb8023049e1aec": [19],
                    "bb54e4fa3906387218be10cff631a7c2": [24],
                    "58ce3b40d384eacd9bad7d916eb8f705": [26],
                    "0327cfb13205ec8512e1c28e438ab43b": [22],
                    "0808e185a52825d22356de2fe00a8f5f": [26],
                    "90bb41c1fbd7c71c05dbc8695320af71": [20],
                    "6c7e43619fe5e6963e8159cc84a28321": [27],
                    "976b121b46db6967854b9c1a6628396b": [27],
                    "9e58d3c72c7006b6f5800b623fbc9818": [23],
                    "ce5f02ac408b7b5c500050190f549736": [31],
                    "bb81456ec79522bf02f34b02762f95e0": [27],
                    "b20800a7faafd2808d6c888577a2ba1d": [28],
                    "33d910f95326c0c7dfe7536d1cfeba51": [32],
                    "61e4931d87cb627df2a19b8bc5819b7b": [23],
                    "3692b23382fd4734215465251290c65b": [26],
                    "eff8910cddcdff62e4741243099240d5": [27],
                    "f67d274202366f6b976414c950ca14bd": [29],
                    "7b3ce20a8e8cf3097bb9df270a7ae63a": [20],
                    "73612fdabd337d5a8279acc30ce22d00": [19],
                    "ad2ad2a70c3e0bf94ddef3f893e92291": [22],
                    "8ce0d477f610ea18e2ea4fbbb46de857": [30],
                    "ff5a6cc031dd2c98b8bccd4766af38c1": [27],
                    "940576153c66af14a949fd19aedd5f5b": [22],
                    "ab05b4fb82e37c8cf5b1ac40d0a37fe9": [21],
                    "57c04d62ada3a102248b48f34c755159": [22],
                    "816b147dcf3305839f723a131b9ad6af": [16],
                    "84000630d1b69a0fe870c94fb26a32bc": [20],
                },
                "details": {
                    "metric_configuration": {
                        "metric_name": "column.distinct_values.count",
                        "domain_kwargs": {"column": "tolls_amount"},
                        "metric_value_kwargs": None,
                    },
                    "num_batches": 36,
                },
            }
        },
        Domain(
            domain_type=MetricDomainTypes.COLUMN,
            domain_kwargs={
                "column": "improvement_surcharge",
            },
            details={
                INFERRED_SEMANTIC_TYPE_KEY: {
                    "improvement_surcharge": SemanticDomainTypes.NUMERIC,
                },
            },
            rule_name="categorical_columns_rule",
        ): {
            "$parameter.column_distinct_values_count": {
                "value": [
                    3,
                    3,
                    3,
                    3,
                    3,
                    3,
                    3,
                    3,
                    3,
                    3,
                    3,
                    3,
                    3,
                    3,
                    3,
                    3,
                    3,
                    3,
                    3,
                    3,
                    3,
                    3,
                    3,
                    3,
                    3,
                    3,
                    3,
                    3,
                    3,
                    3,
                    3,
                    3,
                    3,
                    3,
                    3,
                    3,
                ],
                "attributed_value": {
                    "c92d0679f769ac83fef2bb5eaac5d12a": [3],
                    "562969eaef9c843cb4531aecbc13bbcb": [3],
                    "569a4a80bf434c888593c651dbf2f157": [3],
                    "f6c389dcef63c1f214c30f66b66945c0": [3],
                    "c4fe9afce1cf3e83eb8518a9f5abc754": [3],
                    "e20c38f98b9830a40b851939ca7189d4": [3],
                    "f2e4d3da6556638b55df8ce509b094c2": [3],
                    "44c1b1947c9049e7db62c5320dde4c63": [3],
                    "47157bdaf05a7992473cd699cabaef74": [3],
                    "08085632aff9ce4cebbb8023049e1aec": [3],
                    "bb54e4fa3906387218be10cff631a7c2": [3],
                    "58ce3b40d384eacd9bad7d916eb8f705": [3],
                    "0327cfb13205ec8512e1c28e438ab43b": [3],
                    "0808e185a52825d22356de2fe00a8f5f": [3],
                    "90bb41c1fbd7c71c05dbc8695320af71": [3],
                    "6c7e43619fe5e6963e8159cc84a28321": [3],
                    "976b121b46db6967854b9c1a6628396b": [3],
                    "9e58d3c72c7006b6f5800b623fbc9818": [3],
                    "ce5f02ac408b7b5c500050190f549736": [3],
                    "bb81456ec79522bf02f34b02762f95e0": [3],
                    "b20800a7faafd2808d6c888577a2ba1d": [3],
                    "33d910f95326c0c7dfe7536d1cfeba51": [3],
                    "61e4931d87cb627df2a19b8bc5819b7b": [3],
                    "3692b23382fd4734215465251290c65b": [3],
                    "eff8910cddcdff62e4741243099240d5": [3],
                    "f67d274202366f6b976414c950ca14bd": [3],
                    "7b3ce20a8e8cf3097bb9df270a7ae63a": [3],
                    "73612fdabd337d5a8279acc30ce22d00": [3],
                    "ad2ad2a70c3e0bf94ddef3f893e92291": [3],
                    "8ce0d477f610ea18e2ea4fbbb46de857": [3],
                    "ff5a6cc031dd2c98b8bccd4766af38c1": [3],
                    "940576153c66af14a949fd19aedd5f5b": [3],
                    "ab05b4fb82e37c8cf5b1ac40d0a37fe9": [3],
                    "57c04d62ada3a102248b48f34c755159": [3],
                    "816b147dcf3305839f723a131b9ad6af": [3],
                    "84000630d1b69a0fe870c94fb26a32bc": [3],
                },
                "details": {
                    "metric_configuration": {
                        "metric_name": "column.distinct_values.count",
                        "domain_kwargs": {"column": "improvement_surcharge"},
                        "metric_value_kwargs": None,
                    },
                    "num_batches": 36,
                },
            }
        },
        Domain(
            domain_type=MetricDomainTypes.COLUMN,
            domain_kwargs={
                "column": "congestion_surcharge",
            },
            details={
                INFERRED_SEMANTIC_TYPE_KEY: {
                    "congestion_surcharge": SemanticDomainTypes.NUMERIC,
                },
            },
            rule_name="categorical_columns_rule",
        ): {
            "$parameter.column_distinct_values_count": {
                "value": [
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    1,
                    3,
                    3,
                    4,
                    3,
                    3,
                    3,
                    3,
                    3,
                    3,
                    3,
                    3,
                    4,
                    3,
                    3,
                    3,
                    3,
                    3,
                    3,
                    3,
                    3,
                    3,
                    3,
                    3,
                ],
                "attributed_value": {
                    "c92d0679f769ac83fef2bb5eaac5d12a": [0],
                    "562969eaef9c843cb4531aecbc13bbcb": [0],
                    "569a4a80bf434c888593c651dbf2f157": [0],
                    "f6c389dcef63c1f214c30f66b66945c0": [0],
                    "c4fe9afce1cf3e83eb8518a9f5abc754": [0],
                    "e20c38f98b9830a40b851939ca7189d4": [0],
                    "f2e4d3da6556638b55df8ce509b094c2": [0],
                    "44c1b1947c9049e7db62c5320dde4c63": [0],
                    "47157bdaf05a7992473cd699cabaef74": [0],
                    "08085632aff9ce4cebbb8023049e1aec": [0],
                    "bb54e4fa3906387218be10cff631a7c2": [0],
                    "58ce3b40d384eacd9bad7d916eb8f705": [0],
                    "0327cfb13205ec8512e1c28e438ab43b": [1],
                    "0808e185a52825d22356de2fe00a8f5f": [3],
                    "90bb41c1fbd7c71c05dbc8695320af71": [3],
                    "6c7e43619fe5e6963e8159cc84a28321": [4],
                    "976b121b46db6967854b9c1a6628396b": [3],
                    "9e58d3c72c7006b6f5800b623fbc9818": [3],
                    "ce5f02ac408b7b5c500050190f549736": [3],
                    "bb81456ec79522bf02f34b02762f95e0": [3],
                    "b20800a7faafd2808d6c888577a2ba1d": [3],
                    "33d910f95326c0c7dfe7536d1cfeba51": [3],
                    "61e4931d87cb627df2a19b8bc5819b7b": [3],
                    "3692b23382fd4734215465251290c65b": [3],
                    "eff8910cddcdff62e4741243099240d5": [4],
                    "f67d274202366f6b976414c950ca14bd": [3],
                    "7b3ce20a8e8cf3097bb9df270a7ae63a": [3],
                    "73612fdabd337d5a8279acc30ce22d00": [3],
                    "ad2ad2a70c3e0bf94ddef3f893e92291": [3],
                    "8ce0d477f610ea18e2ea4fbbb46de857": [3],
                    "ff5a6cc031dd2c98b8bccd4766af38c1": [3],
                    "940576153c66af14a949fd19aedd5f5b": [3],
                    "ab05b4fb82e37c8cf5b1ac40d0a37fe9": [3],
                    "57c04d62ada3a102248b48f34c755159": [3],
                    "816b147dcf3305839f723a131b9ad6af": [3],
                    "84000630d1b69a0fe870c94fb26a32bc": [3],
                },
                "details": {
                    "metric_configuration": {
                        "metric_name": "column.distinct_values.count",
                        "domain_kwargs": {"column": "congestion_surcharge"},
                        "metric_value_kwargs": None,
                    },
                    "num_batches": 36,
                },
            }
        },
    }
    return expected_metrics_by_domain


@pytest.fixture
def quentin_expected_rule_based_profiler_configuration() -> Callable:
    def _profiler_config(
        name: str, exclude_column_names: Optional[List[str]] = None
    ) -> RuleBasedProfilerConfig:
        exclude_column_names = exclude_column_names or []
        expected_rule_based_profiler_config = RuleBasedProfilerConfig(
            config_version=1.0,
            name=name,
            variables={"random_seed": None},
            rules={
                "table_rule": {
                    "variables": {
                        "false_positive_rate": 0.05,
                        "quantile_statistic_interpolation_method": "nearest",
                        "quantile_bias_correction": False,
                        "quantile_bias_std_error_ratio_threshold": None,
                        "estimator": "bootstrap",
                        "n_resamples": 9999,
                        "include_estimator_samples_histogram_in_details": False,
                        "truncate_values": {
                            "lower_bound": 0,
                        },
                        "round_decimals": 0,
                    },
                    "domain_builder": {
                        "class_name": "TableDomainBuilder",
                        "module_name": "great_expectations.rule_based_profiler.domain_builder.table_domain_builder",
                    },
                    "parameter_builders": [
                        {
                            "replace_nan_with_zero": True,
                            "name": "table_row_count",
                            "module_name": "great_expectations.rule_based_profiler.parameter_builder.metric_multi_batch_parameter_builder",
                            "single_batch_mode": False,
                            "enforce_numeric_metric": True,
                            "class_name": "MetricMultiBatchParameterBuilder",
                            "reduce_scalar_metric": True,
                            "metric_name": "table.row_count",
                        },
                    ],
                    "expectation_configuration_builders": [
                        {
                            "max_value": "$parameter.table_row_count_range.value[1]",
                            "validation_parameter_builder_configs": [
                                {
                                    "replace_nan_with_zero": True,
                                    "name": "table_row_count_range",
                                    "module_name": "great_expectations.rule_based_profiler.parameter_builder.numeric_metric_range_multi_batch_parameter_builder",
                                    "truncate_values": "$variables.truncate_values",
                                    "enforce_numeric_metric": True,
                                    "n_resamples": "$variables.n_resamples",
                                    "class_name": "NumericMetricRangeMultiBatchParameterBuilder",
                                    "estimator": "$variables.estimator",
                                    "reduce_scalar_metric": True,
                                    "metric_name": "table.row_count",
                                    "metric_multi_batch_parameter_builder_name": "table_row_count",
                                    "metric_domain_kwargs": "$domain.domain_kwargs",
                                    "false_positive_rate": "$variables.false_positive_rate",
                                    "quantile_statistic_interpolation_method": "$variables.quantile_statistic_interpolation_method",
                                    "quantile_bias_correction": "$variables.quantile_bias_correction",
                                    "quantile_bias_std_error_ratio_threshold": "$variables.quantile_bias_std_error_ratio_threshold",
                                    "random_seed": "$variables.random_seed",
                                    "include_estimator_samples_histogram_in_details": "$variables.include_estimator_samples_histogram_in_details",
                                    "round_decimals": "$variables.round_decimals",
                                    "evaluation_parameter_builder_configs": [
                                        {
                                            "enforce_numeric_metric": True,
                                            "replace_nan_with_zero": True,
                                            "name": "table_row_count",
                                            "class_name": "MetricMultiBatchParameterBuilder",
                                            "evaluation_parameter_builder_configs": None,
                                            "metric_value_kwargs": None,
                                            "module_name": "great_expectations.rule_based_profiler.parameter_builder.metric_multi_batch_parameter_builder",
                                            "metric_domain_kwargs": None,
                                            "single_batch_mode": False,
                                            "reduce_scalar_metric": True,
                                            "metric_name": "table.row_count",
                                        }
                                    ],
                                },
                            ],
                            "expectation_type": "expect_table_row_count_to_be_between",
                            "module_name": "great_expectations.rule_based_profiler.expectation_configuration_builder.default_expectation_configuration_builder",
                            "meta": {
                                "profiler_details": "$parameter.table_row_count_range.details"
                            },
                            "class_name": "DefaultExpectationConfigurationBuilder",
                            "min_value": "$parameter.table_row_count_range.value[0]",
                        },
                    ],
                },
                "categorical_columns_rule": {
                    "variables": {
                        "cardinality_limit_mode": "FEW",
                        "mostly": 1.0,
                        "strict_min": False,
                        "strict_max": False,
                        "false_positive_rate": 0.05,
                        "quantile_statistic_interpolation_method": "nearest",
                        "quantile_bias_correction": False,
                        "quantile_bias_std_error_ratio_threshold": None,
                        "estimator": "bootstrap",
                        "n_resamples": 9999,
                        "include_estimator_samples_histogram_in_details": False,
                        "truncate_values": {
                            "lower_bound": 0.0,
                        },
                        "round_decimals": 15,
                    },
                    "domain_builder": {
                        "exclude_column_name_suffixes": ["_id", "_ID"],
                        "module_name": "great_expectations.rule_based_profiler.domain_builder.categorical_column_domain_builder",
                        "allowed_semantic_types_passthrough": ["logic"],
                        "exclude_column_names": sorted(
                            [
                                "id",
                                "ID",
                                "Id",
                            ]
                            + exclude_column_names
                        ),
                        "class_name": "CategoricalColumnDomainBuilder",
                        "cardinality_limit_mode": "$variables.cardinality_limit_mode",
                        "exclude_semantic_types": ["binary", "currency", "identifier"],
                    },
                    "parameter_builders": [
                        {
                            "metric_domain_kwargs": "$domain.domain_kwargs",
                            "replace_nan_with_zero": True,
                            "name": "column_distinct_values_count",
                            "module_name": "great_expectations.rule_based_profiler.parameter_builder.metric_multi_batch_parameter_builder",
                            "enforce_numeric_metric": True,
                            "class_name": "MetricMultiBatchParameterBuilder",
                            "single_batch_mode": False,
                            "reduce_scalar_metric": True,
                            "metric_name": "column.distinct_values.count",
                        },
                    ],
                    "expectation_configuration_builders": [
                        {
                            "min_value": "$parameter.column_distinct_values_count_range.value[0]",
                            "class_name": "DefaultExpectationConfigurationBuilder",
                            "module_name": "great_expectations.rule_based_profiler.expectation_configuration_builder.default_expectation_configuration_builder",
                            "meta": {
                                "profiler_details": "$parameter.column_distinct_values_count_range.details"
                            },
                            "expectation_type": "expect_column_unique_value_count_to_be_between",
                            "max_value": "$parameter.column_distinct_values_count_range.value[1]",
                            "strict_min": "$variables.strict_min",
                            "strict_max": "$variables.strict_max",
                            "column": "$domain.domain_kwargs.column",
                            "validation_parameter_builder_configs": [
                                {
                                    "class_name": "NumericMetricRangeMultiBatchParameterBuilder",
                                    "metric_domain_kwargs": "$domain.domain_kwargs",
                                    "metric_multi_batch_parameter_builder_name": "column_distinct_values_count",
                                    "estimator": "$variables.estimator",
                                    "false_positive_rate": "$variables.false_positive_rate",
                                    "name": "column_distinct_values_count_range",
                                    "round_decimals": "$variables.round_decimals",
                                    "reduce_scalar_metric": True,
                                    "metric_name": "column.distinct_values.count",
                                    "enforce_numeric_metric": True,
                                    "quantile_statistic_interpolation_method": "$variables.quantile_statistic_interpolation_method",
                                    "quantile_bias_correction": "$variables.quantile_bias_correction",
                                    "quantile_bias_std_error_ratio_threshold": "$variables.quantile_bias_std_error_ratio_threshold",
                                    "replace_nan_with_zero": True,
                                    "n_resamples": "$variables.n_resamples",
                                    "module_name": "great_expectations.rule_based_profiler.parameter_builder.numeric_metric_range_multi_batch_parameter_builder",
                                    "include_estimator_samples_histogram_in_details": "$variables.include_estimator_samples_histogram_in_details",
                                    "truncate_values": "$variables.truncate_values",
                                    "evaluation_parameter_builder_configs": [
                                        {
                                            "enforce_numeric_metric": True,
                                            "replace_nan_with_zero": True,
                                            "name": "column_distinct_values_count",
                                            "class_name": "MetricMultiBatchParameterBuilder",
                                            "evaluation_parameter_builder_configs": None,
                                            "metric_value_kwargs": None,
                                            "module_name": "great_expectations.rule_based_profiler.parameter_builder.metric_multi_batch_parameter_builder",
                                            "metric_domain_kwargs": "$domain.domain_kwargs",
                                            "single_batch_mode": False,
                                            "reduce_scalar_metric": True,
                                            "metric_name": "column.distinct_values.count",
                                        }
                                    ],
                                },
                            ],
                        },
                    ],
                },
            },
        )
        return expected_rule_based_profiler_config

    return _profiler_config


@pytest.fixture
def quentin_expected_expectation_suite(
    quentin_expected_rule_based_profiler_configuration,
) -> Callable:
    def _expectation_suite(name: str) -> ExpectationSuite:
        expected_expect_table_row_count_to_be_between_expectation_configuration: ExpectationConfiguration = ExpectationConfiguration(
            **{
                "expectation_type": "expect_table_row_count_to_be_between",
                "kwargs": {
                    "min_value": 10000,
                    "max_value": 10000,
                },
                "meta": {
                    "profiler_details": {
                        "metric_configuration": {
                            "metric_name": "table.row_count",
                            "domain_kwargs": {},
                            "metric_value_kwargs": None,
                        },
                        "num_batches": 36,
                    },
                },
            },
        )

        expected_expect_column_unique_value_count_to_be_between_expectation_configuration_list: List[
            ExpectationConfiguration
        ] = [
            ExpectationConfiguration(
                **{
                    "meta": {
                        "profiler_details": {
                            "metric_configuration": {
                                "metric_name": "column.distinct_values.count",
                                "domain_kwargs": {"column": "passenger_count"},
                                "metric_value_kwargs": None,
                            },
                            "num_batches": 36,
                        }
                    },
                    "expectation_type": "expect_column_unique_value_count_to_be_between",
                    "kwargs": {
                        "strict_max": False,
                        "max_value": 8,
                        "strict_min": False,
                        "column": "passenger_count",
                        "min_value": 7,
                    },
                }
            ),
            ExpectationConfiguration(
                **{
                    "meta": {
                        "profiler_details": {
                            "metric_configuration": {
                                "metric_name": "column.distinct_values.count",
                                "domain_kwargs": {"column": "store_and_fwd_flag"},
                                "metric_value_kwargs": None,
                            },
                            "num_batches": 36,
                        }
                    },
                    "expectation_type": "expect_column_unique_value_count_to_be_between",
                    "kwargs": {
                        "strict_max": False,
                        "max_value": 2,
                        "strict_min": False,
                        "column": "store_and_fwd_flag",
                        "min_value": 2,
                    },
                }
            ),
            ExpectationConfiguration(
                **{
                    "meta": {
                        "profiler_details": {
                            "metric_configuration": {
                                "metric_name": "column.distinct_values.count",
                                "domain_kwargs": {"column": "payment_type"},
                                "metric_value_kwargs": None,
                            },
                            "num_batches": 36,
                        }
                    },
                    "expectation_type": "expect_column_unique_value_count_to_be_between",
                    "kwargs": {
                        "strict_max": False,
                        "max_value": 4,
                        "strict_min": False,
                        "column": "payment_type",
                        "min_value": 4,
                    },
                }
            ),
            ExpectationConfiguration(
                **{
                    "meta": {
                        "profiler_details": {
                            "metric_configuration": {
                                "metric_name": "column.distinct_values.count",
                                "domain_kwargs": {"column": "extra"},
                                "metric_value_kwargs": None,
                            },
                            "num_batches": 36,
                        }
                    },
                    "expectation_type": "expect_column_unique_value_count_to_be_between",
                    "kwargs": {
                        "strict_max": False,
                        "max_value": 15,
                        "strict_min": False,
                        "column": "extra",
                        "min_value": 5,
                    },
                }
            ),
            ExpectationConfiguration(
                **{
                    "meta": {
                        "profiler_details": {
                            "metric_configuration": {
                                "metric_name": "column.distinct_values.count",
                                "domain_kwargs": {"column": "mta_tax"},
                                "metric_value_kwargs": None,
                            },
                            "num_batches": 36,
                        }
                    },
                    "expectation_type": "expect_column_unique_value_count_to_be_between",
                    "kwargs": {
                        "strict_max": False,
                        "max_value": 4,
                        "strict_min": False,
                        "column": "mta_tax",
                        "min_value": 3,
                    },
                }
            ),
            ExpectationConfiguration(
                **{
                    "meta": {
                        "profiler_details": {
                            "metric_configuration": {
                                "metric_name": "column.distinct_values.count",
                                "domain_kwargs": {"column": "tolls_amount"},
                                "metric_value_kwargs": None,
                            },
                            "num_batches": 36,
                        }
                    },
                    "expectation_type": "expect_column_unique_value_count_to_be_between",
                    "kwargs": {
                        "strict_max": False,
                        "max_value": 31,
                        "strict_min": False,
                        "column": "tolls_amount",
                        "min_value": 18,
                    },
                }
            ),
            ExpectationConfiguration(
                **{
                    "meta": {
                        "profiler_details": {
                            "metric_configuration": {
                                "metric_name": "column.distinct_values.count",
                                "domain_kwargs": {"column": "improvement_surcharge"},
                                "metric_value_kwargs": None,
                            },
                            "num_batches": 36,
                        }
                    },
                    "expectation_type": "expect_column_unique_value_count_to_be_between",
                    "kwargs": {
                        "strict_max": False,
                        "max_value": 3,
                        "strict_min": False,
                        "column": "improvement_surcharge",
                        "min_value": 3,
                    },
                }
            ),
            ExpectationConfiguration(
                **{
                    "meta": {
                        "profiler_details": {
                            "metric_configuration": {
                                "metric_name": "column.distinct_values.count",
                                "domain_kwargs": {"column": "congestion_surcharge"},
                                "metric_value_kwargs": None,
                            },
                            "num_batches": 36,
                        }
                    },
                    "expectation_type": "expect_column_unique_value_count_to_be_between",
                    "kwargs": {
                        "strict_max": False,
                        "max_value": 4,
                        "strict_min": False,
                        "column": "congestion_surcharge",
                        "min_value": 0,
                    },
                }
            ),
        ]

        expected_expectation_configurations: List[ExpectationConfiguration] = (
            [
                expected_expect_table_row_count_to_be_between_expectation_configuration,
            ]
            + expected_expect_column_unique_value_count_to_be_between_expectation_configuration_list
        )

        expectation_suite_name: str = "my_suite"

        expected_expectation_suite = ExpectationSuite(
            expectation_suite_name=expectation_suite_name,
        )

        expectation_configuration: ExpectationConfiguration
        for expectation_configuration in expected_expectation_configurations:
            expected_expectation_suite._add_expectation(
                expectation_configuration=expectation_configuration,
                send_usage_event=False,
            )

        expected_expectation_suite_meta: Dict[str, Any] = {
            "citations": [
                {
                    "citation_date": "2019-09-26T13:42:41.000000Z",
                    "profiler_config": quentin_expected_rule_based_profiler_configuration(
                        name=name
                    ).to_json_dict(),
                    "comment": "Created by effective Rule-Based Profiler of VolumeDataAssistant with the configuration included.\n",
                }
            ]
        }

        expected_expectation_suite.meta = expected_expectation_suite_meta

        return expected_expectation_suite

    return _expectation_suite


@pytest.fixture(scope="module")
def bobby_volume_data_assistant_result(
    bobby_columnar_table_multi_batch_probabilistic_data_context: DataContext,
) -> VolumeDataAssistantResult:
    context: DataContext = bobby_columnar_table_multi_batch_probabilistic_data_context

    batch_request: dict = {
        "datasource_name": "taxi_pandas",
        "data_connector_name": "monthly",
        "data_asset_name": "my_reports",
    }

    data_assistant_result: DataAssistantResult = context.assistants.volume.run(
        batch_request=batch_request,
        estimation="flag_outliers",
    )

    return cast(VolumeDataAssistantResult, data_assistant_result)


@pytest.fixture
def bobby_volume_data_assistant_result_usage_stats_enabled(
    bobby_columnar_table_multi_batch_deterministic_data_context: DataContext,
) -> VolumeDataAssistantResult:
    context: DataContext = bobby_columnar_table_multi_batch_deterministic_data_context

    batch_request: dict = {
        "datasource_name": "taxi_pandas",
        "data_connector_name": "monthly",
        "data_asset_name": "my_reports",
    }

    data_assistant_result: DataAssistantResult = context.assistants.volume.run(
        batch_request=batch_request,
        estimation="flag_outliers",
    )

    return cast(VolumeDataAssistantResult, data_assistant_result)


@pytest.fixture
def quentin_explicit_instantiation_result_actual_time(
    quentin_columnar_table_multi_batch_data_context,
    set_consistent_seed_within_numeric_metric_range_multi_batch_parameter_builder,
) -> Tuple[Validator, VolumeDataAssistantResult]:
    context: DataContext = quentin_columnar_table_multi_batch_data_context

    batch_request: dict = {
        "datasource_name": "taxi_pandas",
        "data_connector_name": "monthly",
        "data_asset_name": "my_reports",
    }

    validator: Validator = get_validator_with_expectation_suite(
        data_context=context,
        batch_list=None,
        batch_request=batch_request,
        expectation_suite_name=None,
        expectation_suite=None,
        component_name="volume_data_assistant",
        persist=False,
    )
    assert len(validator.batches) == 36

    data_assistant_name: str = "test_volume_data_assistant"

    data_assistant = VolumeDataAssistant(
        name=data_assistant_name,
        validator=validator,
    )

    data_assistant_result: DataAssistantResult = data_assistant.run()

    return validator, cast(VolumeDataAssistantResult, data_assistant_result)


@pytest.fixture
@freeze_time("09/26/2019 13:42:41")
def quentin_explicit_instantiation_result_frozen_time(
    quentin_columnar_table_multi_batch_data_context,
    set_consistent_seed_within_numeric_metric_range_multi_batch_parameter_builder,
) -> Tuple[Validator, VolumeDataAssistantResult]:
    context: DataContext = quentin_columnar_table_multi_batch_data_context

    batch_request: dict = {
        "datasource_name": "taxi_pandas",
        "data_connector_name": "monthly",
        "data_asset_name": "my_reports",
    }

    validator: Validator = get_validator_with_expectation_suite(
        data_context=context,
        batch_list=None,
        batch_request=batch_request,
        expectation_suite_name=None,
        expectation_suite=None,
        component_name="volume_data_assistant",
        persist=False,
    )
    assert len(validator.batches) == 36

    data_assistant_name: str = "test_volume_data_assistant"

    data_assistant = VolumeDataAssistant(
        name=data_assistant_name,
        validator=validator,
    )

    data_assistant_result: DataAssistantResult = data_assistant.run()

    return validator, cast(VolumeDataAssistantResult, data_assistant_result)


@pytest.fixture
def quentin_implicit_invocation_result_actual_time(
    quentin_columnar_table_multi_batch_data_context,
    set_consistent_seed_within_numeric_metric_range_multi_batch_parameter_builder,
):
    context: DataContext = quentin_columnar_table_multi_batch_data_context

    batch_request: dict = {
        "datasource_name": "taxi_pandas",
        "data_connector_name": "monthly",
        "data_asset_name": "my_reports",
    }

    data_assistant_result: DataAssistantResult = context.assistants.volume.run(
        batch_request=batch_request,
        estimation="flag_outliers",
    )

    return cast(VolumeDataAssistantResult, data_assistant_result)


@pytest.fixture
@freeze_time("09/26/2019 13:42:41")
def quentin_implicit_invocation_result_frozen_time(
    quentin_columnar_table_multi_batch_data_context,
    set_consistent_seed_within_numeric_metric_range_multi_batch_parameter_builder,
):
    context: DataContext = quentin_columnar_table_multi_batch_data_context

    batch_request: dict = {
        "datasource_name": "taxi_pandas",
        "data_connector_name": "monthly",
        "data_asset_name": "my_reports",
    }

    data_assistant_result: DataAssistantResult = context.assistants.volume.run(
        batch_request=batch_request,
        estimation="flag_outliers",
    )

    return cast(VolumeDataAssistantResult, data_assistant_result)


def _assert_quentin_expected_metrics_by_domain_serialized(
    actual_metrics_by_domain: Dict[Domain, Dict[str, ParameterNode]],
    expected_metrics_by_domain: Dict[Domain, Dict[str, ParameterNode]],
):
    """
    # TODO: <Alex>TODO: 12/5/2022</Alex>
    Comparison with raw "numpy.ndarray" objects requires tedius inspections of structures and sub-structures of result.
    For now, remove raw values from test (they are non-serialized version of values access through "PARAMETER_KEY" key).
    """
    domain: Domain
    parameter_values_for_fully_qualified_parameter_names: Dict[str, ParameterNode]
    # noinspection PyTypeChecker
    metrics_by_domain_serialized: Dict[Domain, Dict[str, ParameterNode]] = {
        domain: dict(
            filter(
                lambda element: element[0].find(RAW_PARAMETER_KEY) == (-1),
                parameter_values_for_fully_qualified_parameter_names.items(),
            )
        )
        for domain, parameter_values_for_fully_qualified_parameter_names in actual_metrics_by_domain.items()
    }

    assert metrics_by_domain_serialized == expected_metrics_by_domain


def run_volume_data_assistant_result_jupyter_notebook_with_new_cell(
    context: DataContext,
    new_cell: str,
    implicit: bool,
):
    """
    To set this test up we:
    - create a suite
    - write code (as a string) for creating a VolumeDataAssistantResult
    - add a new cell to the notebook that was passed to this method
    - write both cells to ipynb file

    We then:
    - load the notebook back from disk
    - execute the notebook (Note: this will raise various errors like
      CellExecutionError if any cell in the notebook fails)
    """
    root_dir: str = context.root_directory

    expectation_suite_name: str = "test_suite"
    context.add_or_update_expectation_suite(
        expectation_suite_name=expectation_suite_name
    )

    notebook_path: str = os.path.join(root_dir, f"run_volume_data_assistant.ipynb")

    notebook_code_initialization: str = """
    from typing import Optional, Union

    import uuid

    import great_expectations as gx
    from great_expectations.data_context import AbstractDataContext
    from great_expectations.validator.validator import Validator
    from great_expectations.rule_based_profiler.data_assistant import (
        DataAssistant,
        VolumeDataAssistant,
    )
    from great_expectations.rule_based_profiler.data_assistant_result import DataAssistantResult
    from great_expectations.rule_based_profiler.helpers.util import get_validator_with_expectation_suite
    import great_expectations.exceptions as gx_exceptions

    context = gx.get_context()

    batch_request: dict = {
        "datasource_name": "taxi_pandas",
        "data_connector_name": "monthly",
        "data_asset_name": "my_reports",
    }

    """

    explicit_instantiation_code: str = """
    validator: Validator = get_validator_with_expectation_suite(
        data_context=context,
        batch_list=None,
        batch_request=batch_request,
        expectation_suite_name=None,
        expectation_suite=None,
        component_name="volume_data_assistant",
        persist=False,
    )

    data_assistant = VolumeDataAssistant(
        name="test_volume_data_assistant",
        validator=validator,
    )

    data_assistant_result: DataAssistantResult = data_assistant.run()
    """

    implicit_invocation_code: str = """
    data_assistant_result: DataAssistantResult = context.assistants.volume.run(batch_request=batch_request)
    """

    notebook_code: str
    if implicit:
        notebook_code = notebook_code_initialization + implicit_invocation_code
    else:
        notebook_code = notebook_code_initialization + explicit_instantiation_code

    nb = nbformat.v4.new_notebook()
    nb["cells"] = []
    nb["cells"].append(nbformat.v4.new_code_cell(notebook_code))
    nb["cells"].append(nbformat.v4.new_code_cell(new_cell))

    # Write notebook to path and load it as NotebookNode
    with open(notebook_path, "w") as f:
        nbformat.write(nb, f)

    nb: nbformat.notebooknode.NotebookNode = load_notebook_from_path(
        notebook_path=notebook_path
    )

    # Run notebook
    ep: nbconvert.preprocessors.ExecutePreprocessor = (
        nbconvert.preprocessors.ExecutePreprocessor(timeout=180, kernel_name="python3")
    )
    ep.preprocess(nb, {"metadata": {"path": root_dir}})


@pytest.mark.integration
def test_volume_data_assistant_result_serialization(
    bobby_volume_data_assistant_result: VolumeDataAssistantResult,
) -> None:
    volume_data_assistant_result_as_dict: dict = (
        bobby_volume_data_assistant_result.to_dict()
    )
    assert (
        set(volume_data_assistant_result_as_dict.keys())
        == DataAssistantResult.ALLOWED_KEYS
    )
    assert (
        bobby_volume_data_assistant_result.to_json_dict()
        == volume_data_assistant_result_as_dict
    )
    assert len(bobby_volume_data_assistant_result.profiler_config.rules) == 2


@mock.patch(
    "great_expectations.core.usage_statistics.usage_statistics.UsageStatisticsHandler.emit"
)
@pytest.mark.integration
@pytest.mark.slow  # 1.06s
def test_volume_data_assistant_result_get_expectation_suite(
    mock_emit,
    bobby_volume_data_assistant_result_usage_stats_enabled: VolumeDataAssistantResult,
):
    expectation_suite_name: str = "my_suite"

    suite: ExpectationSuite = (
        bobby_volume_data_assistant_result_usage_stats_enabled.get_expectation_suite(
            expectation_suite_name=expectation_suite_name
        )
    )

    assert suite is not None and len(suite.expectations) > 0

    assert mock_emit.call_count == 1

    # noinspection PyUnresolvedReferences
    actual_events: List[unittest.mock._Call] = mock_emit.call_args_list
    assert (
        actual_events[-1][0][0]["event"]
        == UsageStatsEvents.DATA_ASSISTANT_RESULT_GET_EXPECTATION_SUITE
    )


@pytest.mark.integration
def test_volume_data_assistant_result_batch_id_to_batch_identifier_display_name_map_coverage(
    bobby_volume_data_assistant_result: VolumeDataAssistantResult,
):
    metrics_by_domain: Optional[
        Dict[Domain, Dict[str, ParameterNode]]
    ] = bobby_volume_data_assistant_result.metrics_by_domain

    parameter_values_for_fully_qualified_parameter_names: Dict[str, ParameterNode]
    parameter_node: ParameterNode
    batch_id: str
    assert all(
        bobby_volume_data_assistant_result._batch_id_to_batch_identifier_display_name_map[
            batch_id
        ]
        is not None
        for parameter_values_for_fully_qualified_parameter_names in metrics_by_domain.values()
        for parameter_node in parameter_values_for_fully_qualified_parameter_names.values()
        for batch_id in (
            parameter_node[FULLY_QUALIFIED_PARAMETER_NAME_ATTRIBUTED_VALUE_KEY]
            if FULLY_QUALIFIED_PARAMETER_NAME_ATTRIBUTED_VALUE_KEY in parameter_node
            else {}
        ).keys()
    )


@pytest.mark.integration
@pytest.mark.slow  # 3.72s
def test_volume_data_assistant_get_metrics_and_expectations_using_explicit_instantiation(
    quentin_explicit_instantiation_result_frozen_time,
    quentin_expected_metrics_by_domain,
    quentin_expected_expectation_suite,
    quentin_expected_rule_based_profiler_configuration,
):
    validator: Validator
    data_assistant_result: DataAssistantResult
    validator, data_assistant_result = quentin_explicit_instantiation_result_frozen_time

    data_assistant_name: str = "test_volume_data_assistant"

    expected_expectation_suite: ExpectationSuite = quentin_expected_expectation_suite(
        name=data_assistant_name
    )

    _assert_quentin_expected_metrics_by_domain_serialized(
        actual_metrics_by_domain=data_assistant_result.metrics_by_domain,
        expected_metrics_by_domain=quentin_expected_metrics_by_domain,
    )

    expectation_configuration: ExpectationConfiguration
    for expectation_configuration in data_assistant_result.expectation_configurations:
        if "profiler_details" in expectation_configuration.meta:
            expectation_configuration.meta["profiler_details"].pop(
                "estimation_histogram", None
            )

    assert (
        data_assistant_result.expectation_configurations
        == expected_expectation_suite.expectations
    )

    assert sorted(
        deep_filter_properties_iterable(
            properties=data_assistant_result.profiler_config.to_json_dict(),
            delete_fields={"random_seed"},
        ),
        key=lambda element: element[0],
    ) == sorted(
        deep_filter_properties_iterable(
            properties=quentin_expected_rule_based_profiler_configuration(
                name=data_assistant_name,
            ).to_json_dict(),
            delete_fields={"random_seed"},
        ),
        key=lambda element: element[0],
    )

    data_assistant_result.citation.pop("profiler_config", None)
    expected_expectation_suite.meta["citations"][0].pop("profiler_config", None)

    data_assistant_result.citation.pop("citation_date", None)
    expected_expectation_suite.meta["citations"][0].pop("citation_date", None)
    assert (
        data_assistant_result.citation
        == expected_expectation_suite.meta["citations"][0]
    )

    actual_expectation_suite: ExpectationSuite = (
        data_assistant_result.get_expectation_suite(expectation_suite_name="my_suite")
    )
    actual_expectation_suite.meta.pop("great_expectations_version", None)
    expected_expectation_suite.meta.pop("great_expectations_version", None)
    actual_expectation_suite.meta["citations"][0].pop("citation_date", None)
    expected_expectation_suite.meta["citations"][0].pop("citation_date", None)
    assert actual_expectation_suite == expected_expectation_suite


@freeze_time("09/26/2019 13:42:41")
@pytest.mark.integration
@pytest.mark.slow  # 3.53s
def test_volume_data_assistant_get_metrics_and_expectations_using_implicit_invocation(
    quentin_implicit_invocation_result_frozen_time,
    quentin_expected_metrics_by_domain,
    quentin_expected_expectation_suite,
    quentin_expected_rule_based_profiler_configuration,
):
    data_assistant_result: DataAssistantResult = (
        quentin_implicit_invocation_result_frozen_time
    )

    registered_data_assistant_name: str = "volume_data_assistant"

    expected_expectation_suite: ExpectationSuite = quentin_expected_expectation_suite(
        name=registered_data_assistant_name
    )

    _assert_quentin_expected_metrics_by_domain_serialized(
        actual_metrics_by_domain=data_assistant_result.metrics_by_domain,
        expected_metrics_by_domain=quentin_expected_metrics_by_domain,
    )

    expectation_configuration: ExpectationConfiguration
    for expectation_configuration in data_assistant_result.expectation_configurations:
        if "profiler_details" in expectation_configuration.meta:
            expectation_configuration.meta["profiler_details"].pop(
                "estimation_histogram", None
            )

    assert (
        data_assistant_result.expectation_configurations
        == expected_expectation_suite.expectations
    )

    data_assistant_result_profiler_config_as_json_dict: dict = (
        deep_filter_properties_iterable(
            properties=data_assistant_result.profiler_config.to_json_dict(),
            delete_fields={"random_seed"},
        )
    )
    data_assistant_result_profiler_config_as_json_dict["rules"][
        "categorical_columns_rule"
    ]["domain_builder"]["exclude_semantic_types"] = sorted(
        data_assistant_result_profiler_config_as_json_dict["rules"][
            "categorical_columns_rule"
        ]["domain_builder"]["exclude_semantic_types"]
    )
    quentin_expected_rule_based_profiler_configuration_as_json_dict = (
        deep_filter_properties_iterable(
            properties=quentin_expected_rule_based_profiler_configuration(
                name=registered_data_assistant_name
            ).to_json_dict(),
            delete_fields={"random_seed"},
        )
    )
    assert sorted(
        data_assistant_result_profiler_config_as_json_dict,
        key=lambda element: element[0],
    ) == sorted(
        quentin_expected_rule_based_profiler_configuration_as_json_dict,
        key=lambda element: element[0],
    )

    data_assistant_result.citation.pop("profiler_config", None)
    expected_expectation_suite.meta["citations"][0].pop("profiler_config", None)

    data_assistant_result.citation.pop("citation_date", None)
    expected_expectation_suite.meta["citations"][0].pop("citation_date", None)
    assert (
        data_assistant_result.citation
        == expected_expectation_suite.meta["citations"][0]
    )

    actual_expectation_suite: ExpectationSuite = (
        data_assistant_result.get_expectation_suite(expectation_suite_name="my_suite")
    )
    actual_expectation_suite.meta.pop("great_expectations_version", None)
    expected_expectation_suite.meta.pop("great_expectations_version", None)
    actual_expectation_suite.meta["citations"][0].pop("citation_date", None)
    expected_expectation_suite.meta["citations"][0].pop("citation_date", None)
    assert actual_expectation_suite == expected_expectation_suite


@freeze_time("09/26/2019 13:42:41")
@pytest.mark.integration
@pytest.mark.slow  # 3.03s
def test_volume_data_assistant_get_metrics_and_expectations_using_implicit_invocation_with_domain_type_directives(
    quentin_columnar_table_multi_batch_data_context,
    set_consistent_seed_within_numeric_metric_range_multi_batch_parameter_builder,
    quentin_expected_metrics_by_domain,
    quentin_expected_expectation_suite,
    quentin_expected_rule_based_profiler_configuration,
):
    context: DataContext = quentin_columnar_table_multi_batch_data_context

    batch_request: dict = {
        "datasource_name": "taxi_pandas",
        "data_connector_name": "monthly",
        "data_asset_name": "my_reports",
    }

    exclude_column_names: List[str] = [
        "dropoff_datetime",
        "store_and_fwd_flag",
        "extra",
        "congestion_surcharge",
    ]
    data_assistant_result: DataAssistantResult = context.assistants.volume.run(
        batch_request=batch_request,
        estimation=NumericRangeEstimatorType.FLAG_OUTLIERS,
        exclude_column_names=exclude_column_names,
    )

    column_name: str
    expected_excluded_domains: List[Domain] = [
        Domain(
            domain_type=MetricDomainTypes.COLUMN,
            domain_kwargs={
                "column": column_name,
            },
        )
        for column_name in exclude_column_names
    ]

    domain_key: Domain

    # noinspection PyTypeChecker
    quentin_expected_metrics_by_domain = dict(
        filter(
            lambda element: not any(
                element[0].is_superset(other=domain_key)
                for domain_key in expected_excluded_domains
            ),
            quentin_expected_metrics_by_domain.items(),
        )
    )

    registered_data_assistant_name: str = "volume_data_assistant"

    expected_expectation_suite: ExpectationSuite = quentin_expected_expectation_suite(
        name=registered_data_assistant_name
    )

    _assert_quentin_expected_metrics_by_domain_serialized(
        actual_metrics_by_domain=data_assistant_result.metrics_by_domain,
        expected_metrics_by_domain=quentin_expected_metrics_by_domain,
    )

    expectation_configuration: ExpectationConfiguration

    expected_expectation_suite.expectations = [
        expectation_configuration
        for expectation_configuration in expected_expectation_suite.expectations
        if not (
            expectation_configuration.kwargs
            and expectation_configuration.kwargs.get("column") in exclude_column_names
        )
    ]

    for expectation_configuration in data_assistant_result.expectation_configurations:
        if "profiler_details" in expectation_configuration.meta:
            expectation_configuration.meta["profiler_details"].pop(
                "estimation_histogram", None
            )

    assert (
        data_assistant_result.expectation_configurations
        == expected_expectation_suite.expectations
    )

    data_assistant_result_profiler_config_as_json_dict: dict = (
        deep_filter_properties_iterable(
            properties=data_assistant_result.profiler_config.to_json_dict(),
            delete_fields={"random_seed"},
        )
    )
    quentin_expected_rule_based_profiler_configuration_as_json_dict: dict = (
        deep_filter_properties_iterable(
            properties=quentin_expected_rule_based_profiler_configuration(
                name=registered_data_assistant_name,
                exclude_column_names=exclude_column_names,
            ).to_json_dict(),
            delete_fields={"random_seed"},
        )
    )
    data_assistant_result_profiler_config_as_json_dict["rules"][
        "categorical_columns_rule"
    ]["domain_builder"]["exclude_column_names"] = sorted(
        data_assistant_result_profiler_config_as_json_dict["rules"][
            "categorical_columns_rule"
        ]["domain_builder"]["exclude_column_names"]
    )
    quentin_expected_rule_based_profiler_configuration_as_json_dict["rules"][
        "categorical_columns_rule"
    ]["domain_builder"]["exclude_column_names"] = sorted(
        quentin_expected_rule_based_profiler_configuration_as_json_dict["rules"][
            "categorical_columns_rule"
        ]["domain_builder"]["exclude_column_names"]
    )
    data_assistant_result_profiler_config_as_json_dict["rules"][
        "categorical_columns_rule"
    ]["domain_builder"]["exclude_semantic_types"] = sorted(
        data_assistant_result_profiler_config_as_json_dict["rules"][
            "categorical_columns_rule"
        ]["domain_builder"]["exclude_semantic_types"]
    )
    assert sorted(
        data_assistant_result_profiler_config_as_json_dict,
        key=lambda element: element[0],
    ) == sorted(
        quentin_expected_rule_based_profiler_configuration_as_json_dict,
        key=lambda element: element[0],
    )

    data_assistant_result.citation.pop("profiler_config", None)
    expected_expectation_suite.meta["citations"][0].pop("profiler_config", None)

    data_assistant_result.citation.pop("citation_date", None)
    expected_expectation_suite.meta["citations"][0].pop("citation_date", None)
    assert (
        data_assistant_result.citation
        == expected_expectation_suite.meta["citations"][0]
    )

    actual_expectation_suite: ExpectationSuite = (
        data_assistant_result.get_expectation_suite(expectation_suite_name="my_suite")
    )
    actual_expectation_suite.meta.pop("great_expectations_version", None)
    expected_expectation_suite.meta.pop("great_expectations_version", None)
    actual_expectation_suite.meta["citations"][0].pop("citation_date", None)
    expected_expectation_suite.meta["citations"][0].pop("citation_date", None)
    assert actual_expectation_suite == expected_expectation_suite


@freeze_time("09/26/2019 13:42:41")
@pytest.mark.integration
@pytest.mark.slow  # 3.30s
def test_volume_data_assistant_get_metrics_and_expectations_using_implicit_invocation_with_estimation_directive(
    quentin_columnar_table_multi_batch_data_context,
):
    context: DataContext = quentin_columnar_table_multi_batch_data_context

    batch_request: dict = {
        "datasource_name": "taxi_pandas",
        "data_connector_name": "monthly",
        "data_asset_name": "my_reports",
    }

    data_assistant_result: DataAssistantResult = context.assistants.volume.run(
        batch_request=batch_request,
    )

    rule_config: dict
    assert all(
        [
            rule_config["variables"]["estimator"] == "exact"
            if "estimator" in rule_config["variables"]
            else True
            for rule_config in data_assistant_result.profiler_config.rules.values()
        ]
    )


@pytest.mark.integration
@pytest.mark.slow  # 3.31s
def test_volume_data_assistant_get_metrics_and_expectations_using_implicit_invocation_with_variables_directives(
    quentin_columnar_table_multi_batch_data_context,
):
    context: DataContext = quentin_columnar_table_multi_batch_data_context

    batch_request: dict = {
        "datasource_name": "taxi_pandas",
        "data_connector_name": "monthly",
        "data_asset_name": "my_reports",
    }

    data_assistant_result: DataAssistantResult = context.assistants.volume.run(
        batch_request=batch_request,
        estimation=NumericRangeEstimatorType.FLAG_OUTLIERS,
        categorical_columns_rule={
            "false_positive_rate": 0.1,
        },
    )
    assert (
        data_assistant_result.profiler_config.rules["categorical_columns_rule"][
            "variables"
        ]["false_positive_rate"]
        == 1.0e-1
    )


@pytest.mark.integration
@pytest.mark.slow  # 4.97s
def test_volume_data_assistant_execution_time_within_proper_bounds_using_explicit_instantiation(
    quentin_explicit_instantiation_result_actual_time,
):
    validator: Validator
    data_assistant_result: DataAssistantResult
    validator, data_assistant_result = quentin_explicit_instantiation_result_actual_time

    # Rule-Based Profiler execution time (in seconds) must have non-trivial value.
    assert data_assistant_result.profiler_execution_time > 0.0


@pytest.mark.integration
@pytest.mark.slow  # 3.37s
def test_volume_data_assistant_execution_time_within_proper_bounds_using_implicit_invocation(
    quentin_implicit_invocation_result_actual_time,
):
    data_assistant_result: DataAssistantResult = (
        quentin_implicit_invocation_result_actual_time
    )

    # Rule-Based Profiler execution time (in seconds) must have non-trivial value.
    assert data_assistant_result.profiler_execution_time > 0.0


@pytest.mark.integration
@pytest.mark.slow  # 3.46s
def test_volume_data_assistant_batch_id_order_consistency_in_attributed_metrics_by_domain_using_explicit_instantiation(
    quentin_explicit_instantiation_result_actual_time,
):
    validator: Validator
    data_assistant_result: DataAssistantResult
    validator, data_assistant_result = quentin_explicit_instantiation_result_actual_time
    metrics_by_domain: Optional[
        Dict[Domain, Dict[str, ParameterNode]]
    ] = data_assistant_result.metrics_by_domain

    batch: Batch
    expected_batch_ids: List[str] = [batch.id for batch in validator.batches.values()]

    parameter_values_for_fully_qualified_parameter_names: Dict[str, ParameterNode]
    fully_qualified_parameter_name: str
    parameter_node: ParameterNode
    assert all(
        list(parameter_node[FULLY_QUALIFIED_PARAMETER_NAME_ATTRIBUTED_VALUE_KEY].keys())
        == expected_batch_ids
        for parameter_values_for_fully_qualified_parameter_names in metrics_by_domain.values()
        for fully_qualified_parameter_name, parameter_node in parameter_values_for_fully_qualified_parameter_names.items()
    )


@pytest.mark.integration
@pytest.mark.slow  # 13.77s
def test_volume_data_assistant_plot_descriptive_notebook_execution_fails(
    bobby_columnar_table_multi_batch_probabilistic_data_context,
):
    context: DataContext = bobby_columnar_table_multi_batch_probabilistic_data_context

    new_cell: str = (
        "data_assistant_result.plot_metrics(this_is_not_a_real_parameter=True)"
    )

    with pytest.raises(nbconvert.preprocessors.CellExecutionError):
        run_volume_data_assistant_result_jupyter_notebook_with_new_cell(
            context=context,
            new_cell=new_cell,
            implicit=False,
        )

    with pytest.raises(nbconvert.preprocessors.CellExecutionError):
        run_volume_data_assistant_result_jupyter_notebook_with_new_cell(
            context=context,
            new_cell=new_cell,
            implicit=True,
        )


@pytest.mark.integration
@pytest.mark.slow  # 11.07s
def test_volume_data_assistant_plot_descriptive_notebook_execution(
    bobby_columnar_table_multi_batch_probabilistic_data_context,
):
    context: DataContext = bobby_columnar_table_multi_batch_probabilistic_data_context

    new_cell: str = "data_assistant_result.plot_metrics()"

    run_volume_data_assistant_result_jupyter_notebook_with_new_cell(
        context=context,
        new_cell=new_cell,
        implicit=False,
    )

    run_volume_data_assistant_result_jupyter_notebook_with_new_cell(
        context=context,
        new_cell=new_cell,
        implicit=True,
    )


@pytest.mark.integration
@pytest.mark.slow  # 11.91s
def test_volume_data_assistant_plot_prescriptive_notebook_execution(
    bobby_columnar_table_multi_batch_probabilistic_data_context,
):
    context: DataContext = bobby_columnar_table_multi_batch_probabilistic_data_context

    new_cell: str = "data_assistant_result.plot_expectations_and_metrics()"

    run_volume_data_assistant_result_jupyter_notebook_with_new_cell(
        context=context,
        new_cell=new_cell,
        implicit=False,
    )

    run_volume_data_assistant_result_jupyter_notebook_with_new_cell(
        context=context,
        new_cell=new_cell,
        implicit=True,
    )


@pytest.mark.integration
@pytest.mark.slow  # 11.57s
def test_volume_data_assistant_plot_descriptive_theme_notebook_execution(
    bobby_columnar_table_multi_batch_probabilistic_data_context,
):
    context: DataContext = bobby_columnar_table_multi_batch_probabilistic_data_context

    theme = {"font": "Comic Sans MS"}

    new_cell: str = f"data_assistant_result.plot_metrics(theme={theme})"

    run_volume_data_assistant_result_jupyter_notebook_with_new_cell(
        context=context,
        new_cell=new_cell,
        implicit=False,
    )

    run_volume_data_assistant_result_jupyter_notebook_with_new_cell(
        context=context,
        new_cell=new_cell,
        implicit=True,
    )


@pytest.mark.integration
@pytest.mark.slow  # 12.09s
def test_volume_data_assistant_plot_prescriptive_theme_notebook_execution(
    bobby_columnar_table_multi_batch_probabilistic_data_context,
):
    context: DataContext = bobby_columnar_table_multi_batch_probabilistic_data_context

    theme = {"font": "Comic Sans MS"}

    new_cell: str = (
        f"data_assistant_result.plot_expectations_and_metrics(theme={theme})"
    )

    run_volume_data_assistant_result_jupyter_notebook_with_new_cell(
        context=context,
        new_cell=new_cell,
        implicit=False,
    )

    run_volume_data_assistant_result_jupyter_notebook_with_new_cell(
        context=context,
        new_cell=new_cell,
        implicit=True,
    )


@pytest.mark.integration
def test_volume_data_assistant_plot_returns_proper_dict_repr_of_table_domain_chart(
    bobby_volume_data_assistant_result: VolumeDataAssistantResult,
) -> None:
    plot_result: PlotResult = bobby_volume_data_assistant_result.plot_metrics()

    table_domain_chart: dict = plot_result.charts[0].to_dict()
    assert find_strings_in_nested_obj(table_domain_chart, ["Table Row Count per Batch"])


@pytest.mark.integration
def test_volume_data_assistant_plot_returns_proper_dict_repr_of_column_domain_chart(
    bobby_volume_data_assistant_result: VolumeDataAssistantResult,
) -> None:
    plot_result: PlotResult = bobby_volume_data_assistant_result.plot_metrics()

    column_domain_charts: List[dict] = [p.to_dict() for p in plot_result.charts[1:]]
    assert len(column_domain_charts) == 10  # One for each column present

    columns: List[str] = [
        "VendorID",
        "RatecodeID",
        "passenger_count",
        "store_and_fwd_flag",
        "payment_type",
        "extra",
        "mta_tax",
        "tolls_amount",
        "improvement_surcharge",
        "congestion_surcharge",
    ]
    assert find_strings_in_nested_obj(column_domain_charts, columns)


@pytest.mark.integration
def test_volume_data_assistant_plot_include_column_names_filters_output(
    bobby_volume_data_assistant_result: VolumeDataAssistantResult,
) -> None:
    include_column_names: List[str] = [
        "VendorID",
    ]
    plot_result: PlotResult = bobby_volume_data_assistant_result.plot_metrics(
        include_column_names=include_column_names
    )

    column_domain_charts: List[dict] = [p.to_dict() for p in plot_result.charts[1:]]
    assert len(column_domain_charts) == 1  # Normally 18 without filtering
    assert find_strings_in_nested_obj(column_domain_charts, include_column_names)


@pytest.mark.integration
def test_volume_data_assistant_plot_exclude_column_names_filters_output(
    bobby_volume_data_assistant_result: VolumeDataAssistantResult,
) -> None:
    exclude_column_names: List[str] = ["VendorID", "pickup_datetime"]
    plot_result: PlotResult = bobby_volume_data_assistant_result.plot_metrics(
        exclude_column_names=exclude_column_names
    )

    column_domain_charts: List[dict] = [p.to_dict() for p in plot_result.charts[1:]]
    assert len(column_domain_charts) == 9  # Normally 18 without filtering
    assert not find_strings_in_nested_obj(column_domain_charts, exclude_column_names)


@pytest.mark.integration
def test_volume_data_assistant_plot_include_and_exclude_column_names_raises_error(
    bobby_volume_data_assistant_result: VolumeDataAssistantResult,
) -> None:
    with pytest.raises(ValueError) as e:
        bobby_volume_data_assistant_result.plot_metrics(
            include_column_names=["VendorID"], exclude_column_names=["pickup_datetime"]
        )

    assert "either use `include_column_names` or `exclude_column_names`" in str(e.value)


@pytest.mark.integration
def test_volume_data_assistant_plot_custom_theme_overrides(
    bobby_volume_data_assistant_result: VolumeDataAssistantResult,
) -> None:
    font: str = "Comic Sans MS"
    title_color: str = "#FFA500"
    title_font_size: int = 48
    point_size: int = 1000
    y_axis_label_color: str = "red"
    y_axis_label_angle: int = 180
    x_axis_title_color: str = "brown"

    theme: Dict[str, Any] = {
        "font": font,
        "title": {
            "color": title_color,
            "fontSize": title_font_size,
        },
        "point": {"size": point_size},
        "axisY": {
            "labelColor": y_axis_label_color,
            "labelAngle": y_axis_label_angle,
        },
        "axisX": {"titleColor": x_axis_title_color},
    }
    plot_result: PlotResult = (
        bobby_volume_data_assistant_result.plot_expectations_and_metrics(theme=theme)
    )

    # ensure a config has been added to each chart
    assert all(
        not isinstance(chart.config, alt.utils.schemapi.UndefinedType)
        for chart in plot_result.charts
    )

    # ensure the theme elements were updated for each chart
    assert all(chart.config.font == font for chart in plot_result.charts)
    assert all(
        chart.config.title["color"] == title_color for chart in plot_result.charts
    )
    assert all(
        chart.config.title["fontSize"] == title_font_size
        for chart in plot_result.charts
    )
    assert all(chart.config.point["size"] == point_size for chart in plot_result.charts)
    assert all(
        chart.config.axisY["labelColor"] == y_axis_label_color
        for chart in plot_result.charts
    )
    assert all(
        chart.config.axisY["labelAngle"] == y_axis_label_angle
        for chart in plot_result.charts
    )
    assert all(
        chart.config.axisX["titleColor"] == x_axis_title_color
        for chart in plot_result.charts
    )


@pytest.mark.integration
def test_volume_data_assistant_plot_return_tooltip(
    bobby_volume_data_assistant_result: VolumeDataAssistantResult,
) -> None:
    plot_result: PlotResult = (
        bobby_volume_data_assistant_result.plot_expectations_and_metrics()
    )

    expected_tooltip: List[alt.Tooltip] = [
        alt.Tooltip(
            **{
                "field": "column",
                "format": "",
                "title": "Column",
                "type": AltairDataTypes.NOMINAL.value,
            }
        ),
        alt.Tooltip(
            **{
                "field": "month",
                "format": "",
                "title": "Month",
                "type": AltairDataTypes.ORDINAL.value,
            }
        ),
        alt.Tooltip(
            **{
                "field": "name",
                "format": "",
                "title": "Name",
                "type": AltairDataTypes.ORDINAL.value,
            }
        ),
        alt.Tooltip(
            **{
                "field": "year",
                "format": "",
                "title": "Year",
                "type": AltairDataTypes.ORDINAL.value,
            }
        ),
        alt.Tooltip(
            **{
                "field": "min_value",
                "format": "",
                "title": "Min Value",
                "type": AltairDataTypes.QUANTITATIVE.value,
            }
        ),
        alt.Tooltip(
            **{
                "field": "max_value",
                "format": "",
                "title": "Max Value",
                "type": AltairDataTypes.QUANTITATIVE.value,
            }
        ),
        alt.Tooltip(
            **{
                "field": "strict_min",
                "format": "",
                "title": "Strict Min",
                "type": AltairDataTypes.NOMINAL.value,
            }
        ),
        alt.Tooltip(
            **{
                "field": "strict_max",
                "format": "",
                "title": "Strict Max",
                "type": AltairDataTypes.NOMINAL.value,
            }
        ),
    ]

    single_column_return_chart: alt.LayerChart = plot_result.charts[2]
    layer_1: alt.Chart = single_column_return_chart.layer[1]
    actual_tooltip: List[alt.Tooltip] = layer_1.encoding.tooltip

    for tooltip in expected_tooltip:
        assert tooltip in actual_tooltip


@pytest.mark.integration
@pytest.mark.slow  # 11.63s
def test_volume_data_assistant_metrics_plot_descriptive_non_sequential_notebook_execution(
    bobby_columnar_table_multi_batch_probabilistic_data_context,
):
    context: DataContext = bobby_columnar_table_multi_batch_probabilistic_data_context

    new_cell: str = "data_assistant_result.plot_metrics(sequential=False)"

    run_volume_data_assistant_result_jupyter_notebook_with_new_cell(
        context=context,
        new_cell=new_cell,
        implicit=False,
    )

    run_volume_data_assistant_result_jupyter_notebook_with_new_cell(
        context=context,
        new_cell=new_cell,
        implicit=True,
    )


@pytest.mark.integration
@pytest.mark.slow  # 12.09s
def test_volume_data_assistant_metrics_and_expectations_plot_descriptive_non_sequential_notebook_execution(
    bobby_columnar_table_multi_batch_probabilistic_data_context,
):
    context: DataContext = bobby_columnar_table_multi_batch_probabilistic_data_context

    new_cell: str = (
        "data_assistant_result.plot_expectations_and_metrics(sequential=False)"
    )

    run_volume_data_assistant_result_jupyter_notebook_with_new_cell(
        context=context,
        new_cell=new_cell,
        implicit=False,
    )

    run_volume_data_assistant_result_jupyter_notebook_with_new_cell(
        context=context,
        new_cell=new_cell,
        implicit=True,
    )



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Dict, List, Optional, Union

import nbformat

from great_expectations.core.batch import (
    BatchRequest,
    standardize_batch_request_display_ordering,
)
from great_expectations.render.renderer.v3.suite_edit_notebook_renderer import (
    SuiteEditNotebookRenderer,
)
from great_expectations.util import deep_filter_properties_iterable

if TYPE_CHECKING:
    from great_expectations.data_context.data_context.abstract_data_context import (
        AbstractDataContext,
    )


class SuiteProfileNotebookRenderer(SuiteEditNotebookRenderer):
    def __init__(
        self,
        context: AbstractDataContext,
        expectation_suite_name: str,
        profiler_name: str,
        batch_request: Optional[Union[str, Dict[str, Any]]] = None,
    ) -> None:
        super().__init__(context=context)

        if batch_request is None:
            batch_request = {}

        deep_filter_properties_iterable(
            properties=batch_request,
            inplace=True,
        )
        batch_request = standardize_batch_request_display_ordering(
            batch_request=batch_request
        )

        self._batch_request = batch_request
        self._validator = context.get_validator(
            batch_request=BatchRequest(**batch_request),
            expectation_suite_name=expectation_suite_name,
        )

        self._profiler_name = profiler_name
        self._expectation_suite_name = self._validator.expectation_suite_name

    def render(self, **kwargs) -> nbformat.NotebookNode:
        self._notebook = nbformat.v4.new_notebook()

        self.add_header()

        # TODO: <Alex>Update when RBP replaces UCP permanently.</Alex>
        if self._profiler_name:
            self._add_rule_based_profiler_cells()
        else:
            self._add_onboarding_data_assistant_cells()

        self.add_footer()

        return self._notebook

    def render_to_disk(self, notebook_file_path: str, **kwargs) -> None:
        """
        Render a notebook to disk from an expectation suite.
        """
        self.render()
        self.write_notebook_to_disk(
            notebook=self._notebook, notebook_file_path=notebook_file_path
        )

    # noinspection PyMethodOverriding
    def add_header(self) -> None:
        self.add_markdown_cell(
            markdown=f"""# Initialize a new Expectation Suite by profiling a batch of your data.
This process helps you avoid writing lots of boilerplate when authoring suites by allowing you to select columns and other factors that you care about and letting a profiler write some candidate expectations for you to adjust.

**Expectation Suite Name**: `{self._expectation_suite_name}`
"""
        )

    def add_footer(
        self,
        batch_request: Optional[Union[str, Dict[str, Any]]] = None,
    ) -> None:
        self.add_markdown_cell(
            markdown="""# Save & review your new Expectation Suite

Let's save the draft expectation suite as a JSON file in the
`great_expectations/expectations` directory of your project and rebuild the Data
 Docs site to make it easy to review your new suite."""
        )
        code_cell: str = """\
print(validator.get_expectation_suite(discard_failed_expectations=False))
validator.save_expectation_suite(discard_failed_expectations=False)

checkpoint_config = {
    "class_name": "SimpleCheckpoint",
    "validations": [
        {
            "batch_request": batch_request,
            "expectation_suite_name": expectation_suite_name
        }
    ]
}
checkpoint = SimpleCheckpoint(
    f"{validator.active_batch_definition.data_asset_name}_{expectation_suite_name}",
    context,
    **checkpoint_config
)
checkpoint_result = checkpoint.run()

context.build_data_docs()

validation_result_identifier = checkpoint_result.list_validation_result_identifiers()[0]
context.open_data_docs(resource_identifier=validation_result_identifier)
"""
        self.add_code_cell(code=code_cell, lint=True)
        self.add_markdown_cell(
            markdown=f"""## Next steps
After you review this initial Expectation Suite in Data Docs you
should edit this suite to make finer grained adjustments to the expectations.
This can be done by running `great_expectations suite edit {self._expectation_suite_name}`."""
        )

    def _add_user_configurable_profiler_cells(self) -> None:
        self.add_code_cell(
            code=f"""\
import datetime

import pandas as pd

import great_expectations as gx
import great_expectations.jupyter_ux
from great_expectations.profile.user_configurable_profiler import (
    UserConfigurableProfiler,
)
from great_expectations.core.batch import BatchRequest
from great_expectations.checkpoint import SimpleCheckpoint
from great_expectations.exceptions import DataContextError

context = gx.get_context()

batch_request = {self._batch_request}

expectation_suite_name = "{self._expectation_suite_name}"

validator = context.get_validator(
    batch_request=BatchRequest(**batch_request),
    expectation_suite_name=expectation_suite_name
)
column_names = [f'"{{column_name}}"' for column_name in validator.columns()]
print(f"Columns: {{', '.join(column_names)}}.")
validator.head(n_rows=5, fetch_all=False)
""",
            lint=True,
        )
        self.add_markdown_cell(
            markdown="""\
# Select columns

Select the columns on which you would like to set expectations and those which you would like to ignore.

Great Expectations will choose which expectations might make sense for a column based on the **data type** and **cardinality** of the data in each selected column.

Simply comment out columns that are important and should be included. You can select multiple lines and use a Jupyter
keyboard shortcut to toggle each line: **Linux/Windows**:
`Ctrl-/`, **macOS**: `Cmd-/`
        """
        )
        self._add_available_columns_list()
        self._add_profiler_instructions()
        self.add_code_cell(
            code="""\
profiler = UserConfigurableProfiler(
    profile_dataset=validator,
    excluded_expectations=None,
    ignored_columns=exclude_column_names,
    not_null_only=False,
    primary_or_compound_key=None,
    semantic_types_dict=None,
    table_expectations_only=False,
    value_set_threshold="MANY",
)
suite = profiler.build_suite()
validator.expectation_suite = suite
""",
            lint=True,
        )

    def _add_rule_based_profiler_cells(self) -> None:
        self.add_code_cell(
            code=f"""\
import datetime

import pandas as pd

import great_expectations as gx
import great_expectations.jupyter_ux
from great_expectations.core.batch import BatchRequest
from great_expectations.checkpoint import SimpleCheckpoint
from great_expectations.exceptions import DataContextError

context = gx.get_context()

batch_request = {self._batch_request}

expectation_suite_name = "{self._expectation_suite_name}"

validator = context.get_validator(
    batch_request=BatchRequest(**batch_request),
    expectation_suite_name=expectation_suite_name
)
validator.head(n_rows=5, fetch_all=False)
""",
            lint=True,
        )
        self._add_profiler_instructions()
        self.add_code_cell(
            code=f"""\
result = context.run_profiler_with_dynamic_arguments(
    name="{self._profiler_name}",
    batch_request=batch_request,
)
validator.expectation_suite = result.get_expectation_suite(
    expectation_suite_name=expectation_suite_name
)
""",
            lint=True,
        )

    def _add_onboarding_data_assistant_cells(self) -> None:
        self.add_code_cell(
            code=f"""\
import datetime

import pandas as pd

import great_expectations as gx
import great_expectations.jupyter_ux
from great_expectations.core.batch import BatchRequest
from great_expectations.checkpoint import SimpleCheckpoint
from great_expectations.exceptions import DataContextError

context = gx.get_context()

batch_request = {self._batch_request}

expectation_suite_name = "{self._expectation_suite_name}"

validator = context.get_validator(
    batch_request=BatchRequest(**batch_request),
    expectation_suite_name=expectation_suite_name
)
column_names = [f'"{{column_name}}"' for column_name in validator.columns()]
print(f"Columns: {{', '.join(column_names)}}.")
validator.head(n_rows=5, fetch_all=False)
""",
            lint=True,
        )
        self.add_markdown_cell(
            markdown="""\
# Select columns

Select the columns on which you would like to set expectations and those which you would like to ignore.

Great Expectations will choose which expectations might make sense for a column based on the **data type** and **cardinality** of the data in each selected column.

Simply comment out columns that are important and should be included. You can select multiple lines and use a Jupyter
keyboard shortcut to toggle each line: **Linux/Windows**:
`Ctrl-/`, **macOS**: `Cmd-/`

Other directives are shown (commented out) as examples of the depth of control possible (see documentation for details).
"""
        )
        self._add_available_columns_list()
        self._add_onboarding_data_assistant_instructions()
        self.add_code_cell(
            code="""\
result = context.assistants.onboarding.run(
    batch_request=batch_request,
    exclude_column_names=exclude_column_names,
)
validator.expectation_suite = result.get_expectation_suite(
    expectation_suite_name=expectation_suite_name
)
""",
            lint=True,
        )

    def _add_profiler_instructions(self) -> None:
        self.add_markdown_cell(
            markdown="""# Run the UserConfigurableProfiler

The suites generated here are **not meant to be production suites** -- they are **a starting point to build upon**.

**To get to a production-grade suite, you will definitely want to [edit this
suite](https://docs.greatexpectations.io/en/latest/guides/how_to_guides/creating_and_editing_expectations/how_to_edit_an_expectation_suite_using_a_disposable_notebook.html?utm_source=notebook&utm_medium=profile_based_expectations)
after this initial step gets you started on the path towards what you want.**

This is highly configurable depending on your goals.
You can ignore columns or exclude certain expectations, specify a threshold for creating value set expectations, or even specify semantic types for a given column.
You can find more information about [how to configure this profiler, including a list of the expectations that it uses, here.](https://docs.greatexpectations.io/en/latest/guides/how_to_guides/creating_and_editing_expectations/how_to_create_an_expectation_suite_with_the_user_configurable_profiler.html)

"""
        )

    def _add_onboarding_data_assistant_instructions(self) -> None:
        self.add_markdown_cell(
            markdown="""# Run the OnboardingDataAssistant

The suites generated here are **not meant to be production suites** -- they are **a starting point to build upon**.

**To get to a production-grade suite, you will definitely want to [edit this
suite](https://docs.greatexpectations.io/docs/guides/expectations/create_expectations_overview#editing-a-saved-expectation-suite)
after this initial step gets you started on the path towards what you want.**

This is highly configurable depending on your goals.
You can ignore columns, specify cardinality of categorical columns, configure semantic types for columns, even adjust thresholds and/or different estimator parameters, etc.
You can find more information about OnboardingDataAssistant and other DataAssistant components (please see documentation for the complete set of DataAssistant controls) [how to choose and control the behavior of the DataAssistant tailored to your goals](https://docs.greatexpectations.io/docs/guides/expectations/data_assistants/how_to_create_an_expectation_suite_with_the_onboarding_data_assistant).

Performance considerations:
- Latency: We optimized for an explicit "question/answer" design, which means we issue **lots** of queries. Connection latency will impact performance.
- Data Volume: Small samples of data will often give you a great starting point for understanding the dataset. Consider configuring a sampled asset and profiling a small number of batches.
    """
        )

    def _add_available_columns_list(self) -> None:
        column_names: List[str]
        column_name: str
        column_names = [
            f'    "{column_name}",\n' for column_name in self._validator.columns()
        ]
        code: str = f'exclude_column_names = [\n{"".join(column_names)}]'
        self.add_code_cell(code=code, lint=True)



from ruamel import yaml

import great_expectations as gx
from great_expectations.core.batch import BatchRequest, RuntimeBatchRequest

context = gx.get_context()

# <snippet name="tests/integration/docusaurus/connecting_to_your_data/filesystem/pandas_python_example.py yaml">
datasource_config = {
    "name": "taxi_datasource",
    "class_name": "Datasource",
    "module_name": "great_expectations.datasource",
    "execution_engine": {
        "module_name": "great_expectations.execution_engine",
        "class_name": "PandasExecutionEngine",
    },
    "data_connectors": {
        "default_runtime_data_connector_name": {
            "class_name": "RuntimeDataConnector",
            "module_name": "great_expectations.datasource.data_connector",
            "batch_identifiers": ["default_identifier_name"],
        },
        "default_inferred_data_connector_name": {
            "class_name": "InferredAssetFilesystemDataConnector",
            "base_directory": "<PATH_TO_YOUR_DATA_HERE>",
            "default_regex": {"group_names": ["data_asset_name"], "pattern": "(.*)"},
        },
    },
}
# </snippet>

# Please note this override is only to provide good UX for docs and tests.
# In normal usage you'd set your path directly in the yaml above.
datasource_config["data_connectors"]["default_inferred_data_connector_name"][
    "base_directory"
] = "../data/"

# <snippet name="tests/integration/docusaurus/connecting_to_your_data/filesystem/pandas_python_example.py test_yaml_config">
context.test_yaml_config(yaml.dump(datasource_config))
# </snippet>

# <snippet name="tests/integration/docusaurus/connecting_to_your_data/filesystem/pandas_python_example.py add_datasource">
context.add_datasource(**datasource_config)
# </snippet>

# Here is a RuntimeBatchRequest using a path to a single CSV file
batch_request = RuntimeBatchRequest(
    datasource_name="taxi_datasource",
    data_connector_name="default_runtime_data_connector_name",
    data_asset_name="<YOUR_MEANINGFUL_NAME>",  # This can be anything that identifies this data_asset for you
    runtime_parameters={"path": "<PATH_TO_YOUR_DATA_HERE>"},  # Add your path here.
    batch_identifiers={"default_identifier_name": "default_identifier"},
)

# Please note this override is only to provide good UX for docs and tests.
# In normal usage you'd set your path directly in the BatchRequest above.
batch_request.runtime_parameters["path"] = "./data/yellow_tripdata_sample_2019-01.csv"

context.add_or_update_expectation_suite(expectation_suite_name="test_suite")
validator = context.get_validator(
    batch_request=batch_request, expectation_suite_name="test_suite"
)
print(validator.head())

# NOTE: The following code is only for testing and can be ignored by users.
assert isinstance(validator, gx.validator.validator.Validator)

# Here is a BatchRequest naming a data_asset
batch_request = BatchRequest(
    datasource_name="taxi_datasource",
    data_connector_name="default_inferred_data_connector_name",
    data_asset_name="<YOUR_DATA_ASSET_NAME>",
)

# Please note this override is only to provide good UX for docs and tests.
# In normal usage you'd set your data asset name directly in the BatchRequest above.
batch_request.data_asset_name = "yellow_tripdata_sample_2019-01.csv"

context.add_or_update_expectation_suite(expectation_suite_name="test_suite")
validator = context.get_validator(
    batch_request=batch_request, expectation_suite_name="test_suite"
)
print(validator.head())

# NOTE: The following code is only for testing and can be ignored by users.
assert isinstance(validator, gx.validator.validator.Validator)
assert [ds["name"] for ds in context.list_datasources()] == ["taxi_datasource"]
assert "yellow_tripdata_sample_2019-01.csv" in set(
    context.get_available_data_asset_names()["taxi_datasource"][
        "default_inferred_data_connector_name"
    ]
)



"""Example Script: How to configure a Spark Datasource

This example script is intended for use in documentation on how to configure a Spark Datasource.

Assert statements are included to ensure that if the behaviour shown in this script breaks it will not pass
tests and will be updated.  These statements can be ignored by users.

Comments with the tags `<snippet>` and `</snippet>` are used to ensure that if this script is updated
the snippets that are specified for use in documentation are maintained.  These comments can be ignored by users.

--documentation--
    https://docs.greatexpectations.io/docs/guides/connecting_to_your_data/datasource_configuration/how_to_configure_a_spark_datasource

To run this code as a local test, use the following console command:
```
pytest -v --docs-tests -m integration -k "how_to_configure_a_spark_datasource" tests/integration/test_script_runner.py
```

To validate the snippets in this file, use the following console command:
```
yarn snippet-check ./tests/integration/docusaurus/connecting_to_your_data/datasource_configuration/how_to_configure_a_spark_datasource.py
```
"""

# The following imports are used as part of verifying that all example snippets are consistent.
# Users may disregard them.
from ruamel import yaml

import great_expectations as gx
from tests.integration.docusaurus.connecting_to_your_data.datasource_configuration.datasource_configuration_test_utilities import (
    is_subset,
)
from tests.integration.docusaurus.connecting_to_your_data.datasource_configuration.full_datasource_configurations import (
    get_full_config_spark_configured_datasource_multi_batch,
    get_full_config_spark_configured_datasource_single_batch,
    get_full_config_spark_inferred_datasource_multi_batch,
    get_full_config_spark_inferred_datasource_single_batch,
    get_full_config_spark_runtime_datasource,
    get_partial_config_universal_datasource_config_elements,
)

data_context: gx.DataContext = gx.get_context()


def validate_universal_config_elements():
    """Validates that the 'universal' configuration keys and values are in fact identical to the keys and values
    in all the full Spark configurations.
    """
    universal_elements = get_partial_config_universal_datasource_config_elements()
    is_subset(
        universal_elements, get_full_config_spark_configured_datasource_single_batch()
    )
    is_subset(
        universal_elements, get_full_config_spark_configured_datasource_multi_batch()
    )
    is_subset(
        universal_elements, get_full_config_spark_inferred_datasource_single_batch()
    )
    is_subset(
        universal_elements, get_full_config_spark_inferred_datasource_multi_batch()
    )
    is_subset(universal_elements, get_full_config_spark_runtime_datasource())


# The following methods correspond to the section headings in the how-to guide linked in the module docstring.


def section_5_add_the_spark_execution_engine_to_your_datasource_configuration():
    """Provides and tests the snippets for section 5 of the Spark Datasource configuration guide."""

    # <snippet name="spark datasource_config up to definition of Spark execution engine">
    datasource_config: dict = {
        "name": "my_datasource_name",
        "class_name": "Datasource",
        "module_name": "great_expectations.datasource",
        "execution_engine": {
            "class_name": "SparkDFExecutionEngine",
            "module_name": "great_expectations.execution_engine",
        },
    }
    # </snippet>

    execution_engine_snippet: dict = {
        "name": "my_datasource_name",
        "class_name": "Datasource",
        "module_name": "great_expectations.datasource",
        # <snippet name="define spark execution engine">
        "execution_engine": {
            "class_name": "SparkDFExecutionEngine",
            "module_name": "great_expectations.execution_engine",
        }
        # </snippet>
    }

    assert datasource_config == execution_engine_snippet
    is_subset(
        datasource_config, get_full_config_spark_configured_datasource_single_batch()
    )
    is_subset(
        datasource_config, get_full_config_spark_configured_datasource_multi_batch()
    )
    is_subset(
        datasource_config, get_full_config_spark_inferred_datasource_single_batch()
    )
    is_subset(
        datasource_config, get_full_config_spark_inferred_datasource_multi_batch()
    )
    is_subset(datasource_config, get_full_config_spark_runtime_datasource())


def section_6_add_a_dictionary_as_the_value_of_the_data_connectors_key():
    """Provides and tests the snippets for section 6 of the Spark Datasource configuration guide."""

    # <snippet name="spark datasource_config up to adding an empty data_connectors dictionary">
    datasource_config: dict = {
        "name": "my_datasource_name",
        "class_name": "Datasource",
        "module_name": "great_expectations.datasource",
        "execution_engine": {
            "class_name": "SparkDFExecutionEngine",
            "module_name": "great_expectations.execution_engine",
        },
        "data_connectors": {},
    }
    # </snippet>
    is_subset(
        datasource_config, get_full_config_spark_configured_datasource_single_batch()
    )
    is_subset(
        datasource_config, get_full_config_spark_configured_datasource_multi_batch()
    )
    is_subset(
        datasource_config, get_full_config_spark_inferred_datasource_single_batch()
    )
    is_subset(
        datasource_config, get_full_config_spark_inferred_datasource_multi_batch()
    )
    is_subset(datasource_config, get_full_config_spark_runtime_datasource())


def section_7_configure_your_individual_data_connectors__inferred():
    """Provides and tests the InferredAssetFilesystemDataConnector snippets for section 7 of the
    Spark Datasource configuration guide.

    """
    # <snippet name="spark Datasource configuration up to adding an empty inferred data connector">
    datasource_config: dict = {
        "name": "my_datasource_name",
        "class_name": "Datasource",
        "module_name": "great_expectations.datasource",
        "execution_engine": {
            "class_name": "SparkDFExecutionEngine",
            "module_name": "great_expectations.execution_engine",
        },
        "data_connectors": {"name_of_my_inferred_data_connector": {}},
    }
    # </snippet>
    is_subset(
        datasource_config, get_full_config_spark_inferred_datasource_single_batch()
    )
    is_subset(
        datasource_config, get_full_config_spark_inferred_datasource_multi_batch()
    )

    # <snippet name="inferred spark datasource_config up to adding empty default_regex and batch_spec_passthrough dictionaries">
    datasource_config: dict = {
        "name": "my_datasource_name",
        "class_name": "Datasource",
        "module_name": "great_expectations.datasource",
        "execution_engine": {
            "class_name": "SparkDFExecutionEngine",
            "module_name": "great_expectations.execution_engine",
        },
        "data_connectors": {
            "name_of_my_inferred_data_connector": {
                # <snippet name="inferred data connector define class_name as InferredAssetFilesystemDataConnector">
                "class_name": "InferredAssetFilesystemDataConnector",
                # </snippet>
                "base_directory": "../data",
                "default_regex": {},
                "batch_spec_passthrough": {},
            }
        },
    }
    # </snippet>

    is_subset(
        datasource_config, get_full_config_spark_inferred_datasource_single_batch()
    )
    is_subset(
        datasource_config, get_full_config_spark_inferred_datasource_multi_batch()
    )

    glob_directive: dict = {
        "name": "my_datasource_name",
        "class_name": "Datasource",
        "module_name": "great_expectations.datasource",
        "execution_engine": {
            "class_name": "SparkDFExecutionEngine",
            "module_name": "great_expectations.execution_engine",
        },
        "data_connectors": {
            "name_of_my_inferred_data_connector": {
                "class_name": "InferredAssetFilesystemDataConnector",
                "base_directory": "../data",
                # <snippet name="define glob_directive">
                "glob_directive": "*/*",
                # </snippet>
                "default_regex": {},
                "batch_spec_passthrough": {},
            }
        },
    }
    is_subset(datasource_config, glob_directive)

    # <snippet name="spark inferred datasource_config post glob_directive definition">
    datasource_config: dict = {
        "name": "my_datasource_name",
        "class_name": "Datasource",
        "module_name": "great_expectations.datasource",
        "execution_engine": {
            "class_name": "SparkDFExecutionEngine",
            "module_name": "great_expectations.execution_engine",
        },
        "data_connectors": {
            "name_of_my_inferred_data_connector": {
                "class_name": "InferredAssetFilesystemDataConnector",
                "base_directory": "../data",
                "glob_directive": "*/*",
                "default_regex": {},
                "batch_spec_passthrough": {},
            }
        },
    }
    # </snippet>
    assert datasource_config == glob_directive


def section_7_configure_your_individual_data_connectors__configured():
    """Provides and tests the ConfiguredAssetFilesystemDataConnector snippets for section 7 of the
    Spark Datasource configuration guide.

    """
    # <snippet name="spark Datasource configuration up to adding an empty configured data connector">
    datasource_config: dict = {
        "name": "my_datasource_name",
        "class_name": "Datasource",
        "module_name": "great_expectations.datasource",
        "execution_engine": {
            "class_name": "SparkDFExecutionEngine",
            "module_name": "great_expectations.execution_engine",
        },
        "data_connectors": {"name_of_my_configured_data_connector": {}},
    }
    # </snippet>
    is_subset(
        datasource_config, get_full_config_spark_configured_datasource_single_batch()
    )
    is_subset(
        datasource_config, get_full_config_spark_configured_datasource_multi_batch()
    )

    # <snippet name="configured spark datasource_config up to adding empty assets and batch_spec_passthrough dictionaries">
    datasource_config: dict = {
        "name": "my_datasource_name",
        "class_name": "Datasource",
        "module_name": "great_expectations.datasource",
        "execution_engine": {
            "class_name": "SparkDFExecutionEngine",
            "module_name": "great_expectations.execution_engine",
        },
        "data_connectors": {
            "name_of_my_configured_data_connector": {
                # <snippet name="configured data connector define class_name as ConfiguredAssetFilesystemDataConnector">
                "class_name": "ConfiguredAssetFilesystemDataConnector",
                # </snippet>
                "base_directory": "../data",
                "assets": {},
                "batch_spec_passthrough": {},
            }
        },
    }
    # </snippet>

    is_subset(
        datasource_config, get_full_config_spark_configured_datasource_single_batch()
    )
    is_subset(
        datasource_config, get_full_config_spark_configured_datasource_multi_batch()
    )


def section_7_configure_your_individual_data_connectors__runtime():
    """Provides and tests the RuntimeDataConnector snippets for section 7 of the
    Spark Datasource configuration guide.

    """
    # <snippet name="spark Datasource configuration up to adding an empty runtime data connector">
    datasource_config: dict = {
        "name": "my_datasource_name",
        "class_name": "Datasource",
        "module_name": "great_expectations.datasource",
        "execution_engine": {
            "class_name": "SparkDFExecutionEngine",
            "module_name": "great_expectations.execution_engine",
        },
        "data_connectors": {"name_of_my_runtime_data_connector": {}},
    }
    # </snippet>
    is_subset(datasource_config, get_full_config_spark_runtime_datasource())

    # <snippet name="runtime spark datasource_config up to adding empty batch_identifiers and batch_spec_passthrough dictionaries">
    datasource_config: dict = {
        "name": "my_datasource_name",
        "class_name": "Datasource",
        "module_name": "great_expectations.datasource",
        "execution_engine": {
            "class_name": "SparkDFExecutionEngine",
            "module_name": "great_expectations.execution_engine",
        },
        "data_connectors": {
            "name_of_my_runtime_data_connector": {
                # <snippet name="runtime data connector define class_name as RuntimeDataConnector">
                "class_name": "RuntimeDataConnector",
                # </snippet>
                "batch_spec_passthrough": {},
                "batch_identifiers": [],
            }
        },
    }
    # </snippet>

    is_subset(datasource_config, get_full_config_spark_runtime_datasource())


def section_8_configure_the_values_for_batch_spec_passthrough__universal():
    """Provides and tests the RuntimeDataConnector snippets for section 7 of the
    Spark Datasource configuration guide.

    """
    empty_batch_spec_passthrough: dict = {
        # <snippet name="populate batch_spec_passthrough with empty reader_method and reader_options">
        "batch_spec_passthrough": {"reader_method": "", "reader_options": {}}
        # </snippet>
    }
    smaller_snippets: dict = {
        "batch_spec_passthrough": {
            # <snippet name="populate reader_method with csv">
            "reader_method": "csv",
            # </snippet>
            # <snippet name = "populate reader_options with empty header and inferSchema keys">
            "reader_options": {"header": "", "inferSchema": ""}
            # </snippet>
        }
    }
    even_smaller_snippets: dict = {
        "batch_spec_passthrough": {
            "reader_method": "csv",
            "reader_options": {
                # <snippet name="define header as True for reader_options">
                "header": True,
                # </snippet>
                # <snippet name="define inferSchema as True for reader_options">
                "inferSchema": True
                # </snippet>
            },
        }
    }
    is_subset(empty_batch_spec_passthrough, even_smaller_snippets)
    is_subset(smaller_snippets, even_smaller_snippets)
    is_subset(
        even_smaller_snippets,
        get_full_config_spark_configured_datasource_single_batch()["data_connectors"][
            "name_of_my_configured_data_connector"
        ],
    )
    is_subset(
        even_smaller_snippets,
        get_full_config_spark_configured_datasource_multi_batch()["data_connectors"][
            "name_of_my_configured_data_connector"
        ],
    )
    is_subset(
        even_smaller_snippets,
        get_full_config_spark_inferred_datasource_single_batch()["data_connectors"][
            "name_of_my_inferred_data_connector"
        ],
    )
    is_subset(
        even_smaller_snippets,
        get_full_config_spark_inferred_datasource_multi_batch()["data_connectors"][
            "name_of_my_inferred_data_connector"
        ],
    )
    is_subset(
        even_smaller_snippets,
        get_full_config_spark_runtime_datasource()["data_connectors"][
            "name_of_my_runtime_data_connector"
        ],
    )
    assert (
        even_smaller_snippets["batch_spec_passthrough"]
        == get_full_config_spark_configured_datasource_single_batch()[
            "data_connectors"
        ]["name_of_my_configured_data_connector"]["batch_spec_passthrough"]
    )
    assert (
        even_smaller_snippets["batch_spec_passthrough"]
        == get_full_config_spark_configured_datasource_multi_batch()["data_connectors"][
            "name_of_my_configured_data_connector"
        ]["batch_spec_passthrough"]
    )
    assert (
        even_smaller_snippets["batch_spec_passthrough"]
        == get_full_config_spark_inferred_datasource_single_batch()["data_connectors"][
            "name_of_my_inferred_data_connector"
        ]["batch_spec_passthrough"]
    )
    assert (
        even_smaller_snippets["batch_spec_passthrough"]
        == get_full_config_spark_inferred_datasource_multi_batch()["data_connectors"][
            "name_of_my_inferred_data_connector"
        ]["batch_spec_passthrough"]
    )
    assert (
        even_smaller_snippets["batch_spec_passthrough"]
        == get_full_config_spark_runtime_datasource()["data_connectors"][
            "name_of_my_runtime_data_connector"
        ]["batch_spec_passthrough"]
    )


def section_8_configure_the_values_for_batch_spec_passthrough__inferred():
    # <snippet name="inferred spark datasource_config post batch_spec_passthrough definition">
    datasource_config: dict = {
        "name": "my_datasource_name",  # Preferably name it something relevant
        "class_name": "Datasource",
        "module_name": "great_expectations.datasource",
        "execution_engine": {
            "class_name": "SparkDFExecutionEngine",
            "module_name": "great_expectations.execution_engine",
        },
        "data_connectors": {
            "name_of_my_inferred_data_connector": {
                "class_name": "InferredAssetFilesystemDataConnector",
                "base_directory": "../data",
                "default_regex": {},
                "batch_spec_passthrough": {
                    "reader_method": "csv",
                    "reader_options": {
                        "header": True,
                        "inferSchema": True,
                    },
                },
            }
        },
    }
    # </snippet>
    is_subset(
        datasource_config, get_full_config_spark_inferred_datasource_single_batch()
    )
    is_subset(
        datasource_config, get_full_config_spark_inferred_datasource_multi_batch()
    )


def section_8_configure_the_values_for_batch_spec_passthrough__configured():
    # <snippet name="configured spark datasource_config post batch_spec_passthrough definition">
    datasource_config: dict = {
        "name": "my_datasource_name",  # Preferably name it something relevant
        "class_name": "Datasource",
        "module_name": "great_expectations.datasource",
        "execution_engine": {
            "class_name": "SparkDFExecutionEngine",
            "module_name": "great_expectations.execution_engine",
        },
        "data_connectors": {
            "name_of_my_configured_data_connector": {
                "class_name": "ConfiguredAssetFilesystemDataConnector",
                "base_directory": "../data",
                "assets": {},
                "batch_spec_passthrough": {
                    "reader_method": "csv",
                    "reader_options": {
                        "header": True,
                        "inferSchema": True,
                    },
                },
            }
        },
    }
    # </snippet>
    is_subset(
        datasource_config, get_full_config_spark_configured_datasource_single_batch()
    )
    is_subset(
        datasource_config, get_full_config_spark_configured_datasource_multi_batch()
    )


def section_8_configure_the_values_for_batch_spec_passthrough__runtime():
    # <snippet name="runtime spark datasource_config post batch_spec_passthrough definition">
    datasource_config: dict = {
        "name": "my_datasource_name",  # Preferably name it something relevant
        "class_name": "Datasource",
        "module_name": "great_expectations.datasource",
        "execution_engine": {
            "class_name": "SparkDFExecutionEngine",
            "module_name": "great_expectations.execution_engine",
        },
        "data_connectors": {
            "name_of_my_runtime_data_connector": {
                "class_name": "RuntimeDataConnector",
                "batch_spec_passthrough": {
                    "reader_method": "csv",
                    "reader_options": {
                        "header": True,
                        "inferSchema": True,
                    },
                },
                "batch_identifiers": [],
            }
        },
    }
    # </snippet>
    is_subset(datasource_config, get_full_config_spark_runtime_datasource())


def section_9_configure_your_data_connectors_data_assets__inferred__single_batch():
    single_line_snippets: dict = {
        "data_connectors": {
            "name_of_my_inferred_data_connector": {
                "default_regex": {
                    # <snippet name="inferred spark single batch pattern">
                    "pattern": "(.*)\\.csv",
                    # </snippet>
                    # <snippet name="inferred spark single batch group_names">
                    "group_names": ["data_asset_name"]
                    # </snippet>
                },
            },
        },
    }
    is_subset(
        single_line_snippets, get_full_config_spark_inferred_datasource_single_batch()
    )

    data_connector_snippet: dict = {
        "data_connectors": {
            # <snippet name="inferred spark single batch data connector config">
            "name_of_my_inferred_data_connector": {
                "class_name": "InferredAssetFilesystemDataConnector",
                "base_directory": "../data",
                "default_regex": {
                    "pattern": "(.*)\\.csv",
                    "group_names": ["data_asset_name"],
                },
            },
            # </snippet>
        },
        # </snippet>
    }
    is_subset(
        data_connector_snippet, get_full_config_spark_inferred_datasource_single_batch()
    )

    # <snippet name="full datasource_config for inferred spark single batch datasource">
    datasource_config: dict = {
        "name": "my_datasource_name",  # Preferably name it something relevant
        "class_name": "Datasource",
        "module_name": "great_expectations.datasource",
        "execution_engine": {
            "class_name": "SparkDFExecutionEngine",
            "module_name": "great_expectations.execution_engine",
        },
        "data_connectors": {
            "name_of_my_inferred_data_connector": {
                "class_name": "InferredAssetFilesystemDataConnector",
                "base_directory": "../data",
                "default_regex": {
                    "pattern": "(.*)\\.csv",
                    "group_names": ["data_asset_name"],
                },
                "batch_spec_passthrough": {
                    "reader_method": "csv",
                    "reader_options": {
                        "header": True,
                        "inferSchema": True,
                    },
                },
            }
        },
    }
    # </snippet>
    is_subset(
        datasource_config, get_full_config_spark_inferred_datasource_single_batch()
    )


def section_9_configure_your_data_connectors_data_assets__inferred__multi_batch():
    # <snippet name="full datasource_config for inferred spark multi batch datasource">
    datasource_config: dict = {
        "name": "my_datasource_name",  # Preferably name it something relevant
        "class_name": "Datasource",
        "module_name": "great_expectations.datasource",
        "execution_engine": {
            "class_name": "SparkDFExecutionEngine",
            "module_name": "great_expectations.execution_engine",
        },
        "data_connectors": {
            # <snippet name="inferred spark multi batch data connector config">
            "name_of_my_inferred_data_connector": {
                "class_name": "InferredAssetFilesystemDataConnector",
                "base_directory": "../data",
                "default_regex": {
                    "pattern": "(yellow_tripdata_sample_2020)-(\\d.*)\\.csv",
                    "group_names": ["data_asset_name", "month"],
                },
                "batch_spec_passthrough": {
                    "reader_method": "csv",
                    "reader_options": {
                        "header": True,
                        "inferSchema": True,
                    },
                },
            }
            # </snippet>
        },
    }
    # </snippet>
    is_subset(
        datasource_config, get_full_config_spark_inferred_datasource_multi_batch()
    )


def section_9_configure_your_data_connectors_data_assets__configured__single_batch():
    datasource_config: dict = {
        "data_connectors": {
            "name_of_my_configured_data_connector": {
                "assets": {
                    # <snippet name="empty asset for configured spark single batch datasource">
                    "yellow_tripdata_jan": {}
                    # </snippet>
                }
            }
        }
    }
    is_subset(
        datasource_config, get_full_config_spark_configured_datasource_single_batch()
    )

    # <snippet name="full datasource_config for configured spark single batch datasource">
    datasource_config: dict = {
        "name": "my_datasource_name",  # Preferably name it something relevant
        "class_name": "Datasource",
        "module_name": "great_expectations.datasource",
        "execution_engine": {
            "class_name": "SparkDFExecutionEngine",
            "module_name": "great_expectations.execution_engine",
        },
        "data_connectors": {
            # <snippet name="configured spark single batch data connector config">
            "name_of_my_configured_data_connector": {
                "class_name": "ConfiguredAssetFilesystemDataConnector",
                "base_directory": "../data",
                "assets": {
                    # <snippet name="configured spark single batch asset">
                    "yellow_tripdata_jan": {
                        # <snippet name="configured spark single batch pattern">
                        "pattern": "yellow_tripdata_sample_2020-(01)\\.csv",
                        # </snippet>
                        # <snippet name="configured spark single batch group_names">
                        "group_names": ["month"],
                        # </snippet>
                    },
                    # </snippet>
                },
                "batch_spec_passthrough": {
                    "reader_method": "csv",
                    "reader_options": {
                        "header": True,
                        "inferSchema": True,
                    },
                },
            }
            # </snippet>
        },
    }
    # </snippet>
    is_subset(
        datasource_config, get_full_config_spark_configured_datasource_single_batch()
    )


def section_9_configure_your_data_connectors_data_assets__configured__multi_batch():
    datasource_config: dict = {
        "data_connectors": {
            "name_of_my_configured_data_connector": {
                "assets": {
                    # <snippet name="empty asset for configured spark multi batch datasource">
                    "yellow_tripdata_2020": {}
                    # </snippet>
                }
            }
        }
    }
    is_subset(
        datasource_config, get_full_config_spark_configured_datasource_multi_batch()
    )

    # <snippet name="full datasource_config for configured spark multi batch datasource">
    datasource_config: dict = {
        "name": "my_datasource_name",  # Preferably name it something relevant
        "class_name": "Datasource",
        "module_name": "great_expectations.datasource",
        "execution_engine": {
            "class_name": "SparkDFExecutionEngine",
            "module_name": "great_expectations.execution_engine",
        },
        "data_connectors": {
            # <snippet name="configured spark multi batch data connector config">
            "name_of_my_configured_data_connector": {
                "class_name": "ConfiguredAssetFilesystemDataConnector",
                "base_directory": "../data",
                "assets": {
                    # <snippet name="configured spark multi batch asset">
                    "yellow_tripdata_2020": {
                        # <snippet name="configured spark multi batch pattern">
                        "pattern": "yellow_tripdata_sample_2020-(.*)\\.csv",
                        # </snippet>
                        # <snippet name="configured spark multi batch group_names">
                        "group_names": ["month"],
                        # </snippet>
                    },
                    # </snippet>
                },
                "batch_spec_passthrough": {
                    "reader_method": "csv",
                    "reader_options": {
                        "header": True,
                        "inferSchema": True,
                    },
                },
            }
            # </snippet>
        },
    }
    # </snippet>
    is_subset(
        datasource_config, get_full_config_spark_configured_datasource_multi_batch()
    )


def section_9_configure_your_data_connectors_data_assets__runtime():
    # <snippet name="full datasource_config for runtime spark datasource">
    datasource_config: dict = {
        "name": "my_datasource_name",  # Preferably name it something relevant
        "class_name": "Datasource",
        "module_name": "great_expectations.datasource",
        "execution_engine": {
            "class_name": "SparkDFExecutionEngine",
            "module_name": "great_expectations.execution_engine",
        },
        "data_connectors": {
            "name_of_my_runtime_data_connector": {
                "class_name": "RuntimeDataConnector",
                "batch_spec_passthrough": {
                    "reader_method": "csv",
                    "reader_options": {
                        "header": True,
                        "inferSchema": True,
                    },
                },
                # <snippet name="runtime spark batch_identifiers">
                "batch_identifiers": ["batch_timestamp"],
                # </snippet>
            }
        },
    }
    # </snippet>
    is_subset(datasource_config, get_full_config_spark_runtime_datasource())


def section_10_test_your_configuration_with_test_yaml_config():
    for datasource_config, connector_name, asset_count in (
        (
            get_full_config_spark_runtime_datasource(),
            "name_of_my_runtime_data_connector",
            0,
        ),
        (
            get_full_config_spark_inferred_datasource_single_batch(),
            "name_of_my_inferred_data_connector",
            12,
        ),
        (
            get_full_config_spark_inferred_datasource_multi_batch(),
            "name_of_my_inferred_data_connector",
            1,
        ),
        (
            get_full_config_spark_configured_datasource_single_batch(),
            "name_of_my_configured_data_connector",
            1,
        ),
        (
            get_full_config_spark_configured_datasource_multi_batch(),
            "name_of_my_configured_data_connector",
            1,
        ),
    ):

        test_result = data_context.test_yaml_config(yaml.dump(datasource_config))
        datasource_check = test_result.self_check(max_examples=12)

        # NOTE: The following code is only for testing and can be ignored by users.
        # Assert that there are no data sets -- those get defined in a Batch Request.
        assert (
            datasource_check["data_connectors"][connector_name]["data_asset_count"]
            == asset_count
        ), f"{connector_name} {asset_count} != {datasource_check['data_connectors'][connector_name]['data_asset_count']}"


def section_12_add_your_new_datasource_to_your_datacontext():
    for datasource_config, datasource_name in (
        (
            get_full_config_spark_runtime_datasource(),
            "name_of_my_runtime_data_connector",
        ),
        (
            get_full_config_spark_inferred_datasource_single_batch(),
            "name_of_my_inferred_data_connector",
        ),
        (
            get_full_config_spark_inferred_datasource_multi_batch(),
            "name_of_my_inferred_data_connector",
        ),
        (
            get_full_config_spark_configured_datasource_single_batch(),
            "name_of_my_configured_data_connector",
        ),
        (
            get_full_config_spark_configured_datasource_multi_batch(),
            "name_of_my_configured_data_connector",
        ),
    ):
        # <snippet name="test your spark datasource config with test_yaml_config">
        data_context.test_yaml_config(yaml.dump(datasource_config))
        # </snippet>

        # <snippet name="add your spark datasource to your data_context">
        data_context.add_datasource(**datasource_config)
        # </snippet>

        # <snippet name="add your spark datasource to your data_context only if it does not already exit">
        # add_datasource only if it doesn't already exist in your Data Context
        try:
            data_context.get_datasource(datasource_config["name"])
        except ValueError:
            data_context.add_datasource(**datasource_config)
        else:
            print(
                f"The datasource {datasource_config['name']} already exists in your Data Context!"
            )
        # </snippet>

        assert (
            datasource_name
            in data_context.list_datasources()[0]["data_connectors"].keys()
        ), f"{datasource_name} not in {data_context.list_datasources()}"
        data_context.delete_datasource(datasource_name="my_datasource_name")


validate_universal_config_elements()

section_5_add_the_spark_execution_engine_to_your_datasource_configuration()

section_6_add_a_dictionary_as_the_value_of_the_data_connectors_key()

section_7_configure_your_individual_data_connectors__inferred()
section_7_configure_your_individual_data_connectors__configured()
section_7_configure_your_individual_data_connectors__runtime()

section_8_configure_the_values_for_batch_spec_passthrough__universal()
section_8_configure_the_values_for_batch_spec_passthrough__inferred()
section_8_configure_the_values_for_batch_spec_passthrough__configured()
section_8_configure_the_values_for_batch_spec_passthrough__runtime()

section_9_configure_your_data_connectors_data_assets__inferred__single_batch()
section_9_configure_your_data_connectors_data_assets__inferred__multi_batch()
section_9_configure_your_data_connectors_data_assets__configured__single_batch()
section_9_configure_your_data_connectors_data_assets__configured__multi_batch()
section_9_configure_your_data_connectors_data_assets__runtime()

section_10_test_your_configuration_with_test_yaml_config()

section_12_add_your_new_datasource_to_your_datacontext()



import pathlib

from great_expectations.experimental.datasources.pandas_datasource import (
    PandasDatasource,
    _PandasDatasource,
)
from great_expectations.experimental.datasources.pandas_filesystem_datasource import (
    PandasFilesystemDatasource,
)
from great_expectations.experimental.datasources.postgres_datasource import (
    PostgresDatasource,
)
from great_expectations.experimental.datasources.spark_datasource import (
    SparkFilesystemDatasource,
)
from great_expectations.experimental.datasources.sqlite_datasource import (
    SqliteDatasource,
)

_PANDAS_SCHEMA_VERSION: str = (
    "1.3.5"  # this is the version schemas we generated for. Update as needed
)
_SCHEMAS_DIR = pathlib.Path(__file__).parent / "schemas"



from typing import Union

import pytest
from ruamel.yaml import YAML

try:
    pyspark = pytest.importorskip("pyspark")
except ImportError:
    pyspark = None

from great_expectations import DataContext
from great_expectations.data_context.util import instantiate_class_from_config
from great_expectations.datasource import BaseDatasource, LegacyDatasource

yaml = YAML()


@pytest.fixture
def data_source_config_with_aws_glue_catalog_data_connectors():
    config: str = f"""
    class_name: Datasource
    execution_engine:
        class_name: SparkDFExecutionEngine
    data_connectors:
        glue_configured_connector:
            class_name: ConfiguredAssetAWSGlueDataCatalogDataConnector
            partitions:
                - PClass
                - SexCode
            assets:
                db_test.tb_titanic:
                    table_name: tb_titanic_with_partitions
                    database_name: db_test
                    splitter_method: split_on_column_value
                    splitter_kwargs:
                        column_name: PClass
                    partitions:
                        - SexCode
                asset2:
                    table_name: tb_titanic_with_partitions
                    database_name: db_test
                    sampling_method: _sample_using_random
                    sampling_kwargs:
                        p: 0.5
                        seed: 0
                asset3:
                    table_name: tb_titanic_without_partitions
                    database_name: db_test
        glue_inferred_connector:
            class_name: InferredAssetAWSGlueDataCatalogDataConnector
    """
    return yaml.load(config)


@pytest.mark.integration
def test_instantiation_from_config(
    glue_titanic_catalog, data_source_config_with_aws_glue_catalog_data_connectors
):
    my_data_source = instantiate_class_from_config(
        data_source_config_with_aws_glue_catalog_data_connectors,
        config_defaults={"module_name": "great_expectations.datasource"},
        runtime_environment={"name": "my_spark_datasource"},
    )

    report: dict = my_data_source.self_check()
    report["execution_engine"].pop("spark_config")
    assert report == {
        "execution_engine": {
            "caching": True,
            "module_name": "great_expectations.execution_engine.sparkdf_execution_engine",
            "class_name": "SparkDFExecutionEngine",
            "persist": True,
            "azure_options": {},
        },
        "data_connectors": {
            "count": 2,
            "glue_configured_connector": {
                "class_name": "ConfiguredAssetAWSGlueDataCatalogDataConnector",
                "data_asset_count": 3,
                "example_data_asset_names": ["asset2", "asset3", "db_test.tb_titanic"],
                "data_assets": {
                    "asset2": {
                        "batch_definition_count": 6,
                        "example_data_references": [
                            {"PClass": "1st", "SexCode": "0"},
                            {"PClass": "1st", "SexCode": "1"},
                            {"PClass": "2nd", "SexCode": "0"},
                        ],
                    },
                    "asset3": {
                        "batch_definition_count": 1,
                        "example_data_references": [{}],
                    },
                    "db_test.tb_titanic": {
                        "batch_definition_count": 2,
                        "example_data_references": [
                            {"SexCode": "0"},
                            {"SexCode": "1"},
                        ],
                    },
                },
                "unmatched_data_reference_count": 0,
                "example_unmatched_data_references": [],
            },
            "glue_inferred_connector": {
                "class_name": "InferredAssetAWSGlueDataCatalogDataConnector",
                "data_asset_count": 2,
                "example_data_asset_names": [
                    "db_test.tb_titanic_with_partitions",
                    "db_test.tb_titanic_without_partitions",
                ],
                "data_assets": {
                    "db_test.tb_titanic_with_partitions": {
                        "batch_definition_count": 6,
                        "example_data_references": [
                            {"PClass": "1st", "SexCode": "0"},
                            {"PClass": "1st", "SexCode": "1"},
                            {"PClass": "2nd", "SexCode": "0"},
                        ],
                    },
                    "db_test.tb_titanic_without_partitions": {
                        "batch_definition_count": 1,
                        "example_data_references": [{}],
                    },
                },
                "unmatched_data_reference_count": 0,
                "example_unmatched_data_references": [],
            },
        },
    }


@pytest.mark.integration
def test_instantiation_from_datasource(
    glue_titanic_catalog,
    empty_data_context,
    data_source_config_with_aws_glue_catalog_data_connectors,
):
    # This is a basic integration test demonstrating a Datasource containing a SQL data_connector.
    # It tests that splitter configurations can be saved and loaded to great_expectations.yml by performing a
    # round-trip to the configuration.
    context: DataContext = empty_data_context

    context.add_datasource(
        name="my_datasource",
        **data_source_config_with_aws_glue_catalog_data_connectors,
    )
    datasource: Union[LegacyDatasource, BaseDatasource, None] = context.get_datasource(
        datasource_name="my_datasource"
    )
    report: dict = datasource.self_check()
    report["execution_engine"].pop("spark_config")
    assert report == {
        "execution_engine": {
            "caching": True,
            "module_name": "great_expectations.execution_engine.sparkdf_execution_engine",
            "class_name": "SparkDFExecutionEngine",
            "persist": True,
            "azure_options": {},
        },
        "data_connectors": {
            "count": 2,
            "glue_configured_connector": {
                "class_name": "ConfiguredAssetAWSGlueDataCatalogDataConnector",
                "data_asset_count": 3,
                "example_data_asset_names": ["asset2", "asset3", "db_test.tb_titanic"],
                "data_assets": {
                    "asset2": {
                        "batch_definition_count": 6,
                        "example_data_references": [
                            {"PClass": "1st", "SexCode": "0"},
                            {"PClass": "1st", "SexCode": "1"},
                            {"PClass": "2nd", "SexCode": "0"},
                        ],
                    },
                    "asset3": {
                        "batch_definition_count": 1,
                        "example_data_references": [{}],
                    },
                    "db_test.tb_titanic": {
                        "batch_definition_count": 2,
                        "example_data_references": [
                            {"SexCode": "0"},
                            {"SexCode": "1"},
                        ],
                    },
                },
                "unmatched_data_reference_count": 0,
                "example_unmatched_data_references": [],
            },
            "glue_inferred_connector": {
                "class_name": "InferredAssetAWSGlueDataCatalogDataConnector",
                "data_asset_count": 2,
                "example_data_asset_names": [
                    "db_test.tb_titanic_with_partitions",
                    "db_test.tb_titanic_without_partitions",
                ],
                "data_assets": {
                    "db_test.tb_titanic_with_partitions": {
                        "batch_definition_count": 6,
                        "example_data_references": [
                            {"PClass": "1st", "SexCode": "0"},
                            {"PClass": "1st", "SexCode": "1"},
                            {"PClass": "2nd", "SexCode": "0"},
                        ],
                    },
                    "db_test.tb_titanic_without_partitions": {
                        "batch_definition_count": 1,
                        "example_data_references": [{}],
                    },
                },
                "unmatched_data_reference_count": 0,
                "example_unmatched_data_references": [],
            },
        },
    }
